#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: nameserver_info
short_description: Look up nameservers for a DNS name
version_added: 2.6.0
description:
    - Retrieve all nameservers that are responsible for a DNS name.
extends_documentation_fragment:
    - community.dns.attributes
    - community.dns.attributes.info_module
author:
    - Felix Fontein (@felixfontein)
options:
    name:
        description:
            - A list of DNS names whose nameservers to retrieve.
        required: true
        type: list
        elements: str
    resolve_addresses:
        description:
            - Whether to resolve the nameserver names to IP addresses.
        type: bool
        default: false
    query_retry:
        description:
            - Number of retries for DNS query timeouts.
        type: int
        default: 3
    query_timeout:
        description:
            - Timeout per DNS query in seconds.
        type: float
        default: 10
    always_ask_default_resolver:
        description:
            - When set to V(true) (default), will use the default resolver to find the authoritative nameservers
              of a subzone. See O(server) for how to configure the default resolver.
            - When set to V(false), will use the authoritative nameservers of the parent zone to find the
              authoritative nameservers of a subzone. This only makes sense when the nameservers were recently
              changed and have not yet propagated.
        type: bool
        default: true
    servfail_retries:
        description:
            - How often to retry on SERVFAIL errors.
        type: int
        default: 0
    server:
        description:
            - The DNS server(s) to use to look up the result. Must be a list of one or more IP addresses.
            - By default, the system's standard resolver is used.
        type: list
        elements: str
        version_added: 2.7.0
requirements:
    - dnspython >= 1.15.0 (maybe older versions also work)
'''

EXAMPLES = r'''
- name: Retrieve name servers of two DNS names
  community.dns.nameserver_info:
    name:
      - www.example.com
      - example.org
  register: result

- name: Show nameservers for www.example.com
  ansible.builtin.debug:
    msg: '{{ result.results[0].nameserver }}'
'''

RETURN = r'''
results:
    description:
        - Information on the nameservers for every DNS name provided in O(name).
    returned: always
    type: list
    elements: dict
    contains:
        name:
            description:
                - The DNS name this entry is for.
            returned: always
            type: str
            sample: www.example.com
        nameservers:
            description:
                - A list of nameservers for this DNS name.
            returned: success
            type: list
            elements: str
            sample:
                - ns1.example.com
                - ns2.example.com
    sample:
        - name: www.example.com
          nameservers:
            - ns1.example.com
            - ns2.example.com
        - name: example.org
          nameservers:
            - ns1.example.org
            - ns2.example.org
            - ns3.example.org
'''

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.dns.plugins.module_utils.resolver import (
    ResolveDirectlyFromNameServers,
    assert_requirements_present,
    guarded_run,
)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(required=True, type='list', elements='str'),
            resolve_addresses=dict(type='bool', default=False),
            query_retry=dict(type='int', default=3),
            query_timeout=dict(type='float', default=10),
            always_ask_default_resolver=dict(type='bool', default=True),
            servfail_retries=dict(type='int', default=0),
            server=dict(type='list', elements='str'),
        ),
        supports_check_mode=True,
    )
    assert_requirements_present(module)

    names = module.params['name']
    resolve_addresses = module.params['resolve_addresses']

    resolver = ResolveDirectlyFromNameServers(
        timeout=module.params['query_timeout'],
        timeout_retries=module.params['query_retry'],
        servfail_retries=module.params['servfail_retries'],
        always_ask_default_resolver=module.params['always_ask_default_resolver'],
        server_addresses=module.params['server'],
    )
    results = [None] * len(names)
    for index, name in enumerate(names):
        results[index] = {
            'name': name,
        }

    def f():
        for index, name in enumerate(names):
            results[index]['nameservers'] = sorted(resolver.resolve_nameservers(name, resolve_addresses=resolve_addresses))

    guarded_run(f, module, generate_additional_results=lambda: dict(results=results))
    module.exit_json(results=results)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


__metaclass__ = type

# Copyright: (c) 2017, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)


DOCUMENTATION = """
module: nxos_logging
author: Trishna Guha (@trishnaguha)
short_description: Manage logging on network devices
notes:
- Limited Support for Cisco MDS
description:
- This module provides declarative management of logging on Cisco NX-OS devices.
version_added: 1.0.0
deprecated:
  alternative: nxos_logging_global
  why: Updated module released with more functionality.
  removed_at_date: '2023-08-01'
options:
  dest:
    description:
    - Destination of the logs.
    choices:
    - console
    - logfile
    - module
    - monitor
    - server
    type: str
  remote_server:
    description:
    - Hostname or IP Address for remote logging (when dest is 'server').
    type: str
  use_vrf:
    description:
    - VRF to be used while configuring remote logging (when dest is 'server').
    type: str
  interface:
    description:
    - Interface to be used while configuring source-interface for logging (e.g., 'Ethernet1/2',
      'mgmt0')
    type: str
  name:
    description:
    - If value of C(dest) is I(logfile) it indicates file-name.
    type: str
  facility:
    description:
    - Facility name for logging.
    type: str
  dest_level:
    description:
    - Set logging severity levels.
    aliases:
    - level
    type: int
  facility_level:
    description:
    - Set logging severity levels for facility based log messages.
    type: int
  aggregate:
    description: List of logging definitions.
    type: list
    elements: dict
  state:
    description:
    - State of the logging configuration.
    default: present
    choices:
    - present
    - absent
    type: str
  event:
    description:
    - Link/trunk enable/default interface configuration logging
    choices:
    - link-enable
    - link-default
    - trunk-enable
    - trunk-default
    type: str
  interface_message:
    description:
    - Add interface description to interface syslogs. Does not work with version 6.0
      images using nxapi as a transport.
    choices:
    - add-interface-description
    type: str
  file_size:
    description:
    - Set logfile size
    type: int
  facility_link_status:
    description:
    - Set logging facility ethpm link status. Not idempotent with version 6.0 images.
    choices:
    - link-down-notif
    - link-down-error
    - link-up-notif
    - link-up-error
    type: str
  timestamp:
    description:
    - Set logging timestamp format
    choices:
    - microseconds
    - milliseconds
    - seconds
    type: str
  purge:
    description:
    - Remove any switch logging configuration that does not match what has been configured
      Not supported for ansible_connection local. All nxos_logging tasks must use
      the same ansible_connection type.
    type: bool
    default: false
extends_documentation_fragment:
- cisco.nxos.nxos
"""

EXAMPLES = """
- name: configure console logging with level
  cisco.nxos.nxos_logging:
    dest: console
    level: 2
    state: present
- name: remove console logging configuration
  cisco.nxos.nxos_logging:
    dest: console
    level: 2
    state: absent
- name: configure file logging with level
  cisco.nxos.nxos_logging:
    dest: logfile
    name: testfile
    dest_level: 3
    state: present
- name: Configure logging logfile with size
  cisco.nxos.nxos_logging:
    dest: logfile
    name: testfile
    dest_level: 3
    file_size: 16384
- name: configure facility level logging
  cisco.nxos.nxos_logging:
    facility: daemon
    facility_level: 0
    state: present
- name: remove facility level logging
  cisco.nxos.nxos_logging:
    facility: daemon
    facility_level: 0
    state: absent
- name: Configure Remote Logging
  cisco.nxos.nxos_logging:
    dest: server
    remote_server: test-syslogserver.com
    facility: auth
    facility_level: 1
    use_vrf: management
    state: present
- name: Configure Source Interface for Logging
  cisco.nxos.nxos_logging:
    interface: mgmt0
    state: present
- name: Purge nxos_logging configuration not managed by this playbook
  cisco.nxos.nxos_logging:
    purge: true
- name: Configure logging timestamp
  cisco.nxos.nxos_logging:
    timestamp: milliseconds
    state: present
- name: Configure logging facility ethpm link status
  cisco.nxos.nxos_logging:
    facility: ethpm
    facility_link_status: link-up-notif
    state: present
- name: Configure logging message ethernet description
  cisco.nxos.nxos_logging:
    interface_message: add-interface-description
    state: present
- name: Configure logging event link enable
  cisco.nxos.nxos_logging:
    event: link-enable
    state: present
- name: Configure logging using aggregate
  cisco.nxos.nxos_logging:
    aggregate:
    - {dest: console, dest_level: 2}
    - {dest: logfile, dest_level: 2, name: testfile}
    - {facility: daemon, facility_level: 0}
    state: present
"""

RETURN = """
commands:
  description: The list of configuration mode commands to send to the device
  returned: always
  type: list
  sample:
    - logging console 2
    - logging logfile testfile 3
    - logging level daemon 0
"""

import copy
import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import (
    get_config,
    load_config,
    normalize_interface,
    read_module_context,
    run_commands,
    save_module_context,
)


STATIC_CLI = {
    "link-enable": "logging event link-status enable",
    "link-default": "logging event link-status default",
    "trunk-enable": "logging event trunk-status enable",
    "trunk-default": "logging event trunk-status default",
    "microseconds": "logging timestamp microseconds",
    "milliseconds": "logging timestamp milliseconds",
    "seconds": "logging timestamp seconds",
    "link-up-error": "link-up error",
    "link-up-notif": "link-up notif",
    "link-down-error": "link-down error",
    "link-down-notif": "link-down notif",
    "add-interface-description": "logging message interface type ethernet description",
}

DEFAULT_LOGGING_LEVEL = {
    0: [],
    1: [],
    2: ["pktmgr"],
    3: ["adjmgr", "arp", "icmpv6", "l2rib", "netstack"],
    4: [],
    5: ["mrib", "m6rib"],
    6: [],
    7: [],
}

DEST_GROUP = ["console", "logfile", "module", "monitor", "server"]


def map_obj_to_commands(module, updates):
    commands = list()
    want, have = updates

    for w in want:
        state = w["state"]
        del w["state"]

        if state == "absent" and w in have:
            if w["facility"] is not None:
                if (
                    not w["dest"]
                    and not w["facility_link_status"]
                    and w["facility"] not in DEFAULT_LOGGING_LEVEL[int(w["facility_level"])]
                ):
                    commands.append(
                        "no logging level {0} {1}".format(w["facility"], w["facility_level"]),
                    )

                if w["facility_link_status"] and w["facility"] in ("ethpm"):
                    commands.append(
                        "no logging level {0} {1}".format(
                            w["facility"],
                            STATIC_CLI[w["facility_link_status"]],
                        ),
                    )

            if w["name"] is not None:
                commands.append("no logging logfile")

            if w["dest"] in ("console", "module", "monitor"):
                commands.append("no logging {0}".format(w["dest"]))

            if w["dest"] == "server":
                commands.append("no logging server {0}".format(w["remote_server"]))

            if w["interface"]:
                commands.append("no logging source-interface")

            if w["event"] and w["event"] in STATIC_CLI:
                commands.append("no " + STATIC_CLI[w["event"]])

            if w["message"] and w["message"] in STATIC_CLI:
                commands.append("no " + STATIC_CLI[w["message"]])

            if w["timestamp"] and w["timestamp"] in STATIC_CLI:
                commands.append("no " + STATIC_CLI[w["timestamp"]])

        if state == "present" and w not in have:
            if w["facility"] is None:
                if w["dest"]:
                    if w["dest"] not in ("logfile", "server"):
                        commands.append("logging {0} {1}".format(w["dest"], w["dest_level"]))

                    elif w["dest"] == "logfile":
                        if w["file_size"]:
                            commands.append(
                                "logging logfile {0} {1} size {2}".format(
                                    w["name"],
                                    w["dest_level"],
                                    w["file_size"],
                                ),
                            )
                        else:
                            commands.append(
                                "logging logfile {0} {1}".format(w["name"], w["dest_level"]),
                            )

                    elif w["dest"] == "server":
                        if w["facility_level"]:
                            if w["use_vrf"]:
                                commands.append(
                                    "logging server {0} {1} use-vrf {2}".format(
                                        w["remote_server"],
                                        w["facility_level"],
                                        w["use_vrf"],
                                    ),
                                )
                            else:
                                commands.append(
                                    "logging server {0} {1}".format(
                                        w["remote_server"],
                                        w["facility_level"],
                                    ),
                                )

                        else:
                            if w["use_vrf"]:
                                commands.append(
                                    "logging server {0} use-vrf {1}".format(
                                        w["remote_server"],
                                        w["use_vrf"],
                                    ),
                                )
                            else:
                                commands.append("logging server {0}".format(w["remote_server"]))

            if w["facility"]:
                if w["dest"] == "server":
                    if w["facility_level"]:
                        if w["use_vrf"]:
                            commands.append(
                                "logging server {0} {1} facility {2} use-vrf {3}".format(
                                    w["remote_server"],
                                    w["facility_level"],
                                    w["facility"],
                                    w["use_vrf"],
                                ),
                            )
                        else:
                            commands.append(
                                "logging server {0} {1} facility {2}".format(
                                    w["remote_server"],
                                    w["facility_level"],
                                    w["facility"],
                                ),
                            )
                    else:
                        if w["use_vrf"]:
                            commands.append(
                                "logging server {0} facility {1} use-vrf {2}".format(
                                    w["remote_server"],
                                    w["facility"],
                                    w["use_vrf"],
                                ),
                            )
                        else:
                            commands.append(
                                "logging server {0} facility {1}".format(
                                    w["remote_server"],
                                    w["facility"],
                                ),
                            )
                else:
                    if w["facility_link_status"]:
                        commands.append(
                            "logging level {0} {1}".format(
                                w["facility"],
                                STATIC_CLI[w["facility_link_status"]],
                            ),
                        )
                    else:
                        if not match_facility_default(module, w["facility"], w["facility_level"]):
                            commands.append(
                                "logging level {0} {1}".format(w["facility"], w["facility_level"]),
                            )

            if w["interface"]:
                commands.append(
                    "logging source-interface {0} {1}".format(*split_interface(w["interface"])),
                )

            if w["event"] and w["event"] in STATIC_CLI:
                commands.append(STATIC_CLI[w["event"]])

            if w["message"] and w["message"] in STATIC_CLI:
                commands.append(STATIC_CLI[w["message"]])

            if w["timestamp"] and w["timestamp"] in STATIC_CLI:
                commands.append(STATIC_CLI[w["timestamp"]])

    return commands


def match_facility_default(module, facility, want_level):
    """Check wanted facility to see if it matches current device default"""

    matches_default = False
    # Sample output from show logging level command
    # Facility        Default Severity        Current Session Severity
    # --------        ----------------        ------------------------
    # bfd                     5                       5
    #
    # 0(emergencies)          1(alerts)       2(critical)
    # 3(errors)               4(warnings)     5(notifications)
    # 6(information)          7(debugging)

    regexl = r"\S+\s+(\d+)\s+(\d+)"
    cmd = {
        "command": "show logging level {0}".format(facility),
        "output": "text",
    }
    facility_data = run_commands(module, cmd)
    for line in facility_data[0].split("\n"):
        mo = re.search(regexl, line)
        if mo and int(mo.group(1)) == int(want_level) and int(mo.group(2)) == int(want_level):
            matches_default = True

    return matches_default


def split_interface(interface):
    match = re.search(r"(\D+)(\S*)", interface, re.M)
    if match:
        return match.group(1), match.group(2)


def parse_facility_link_status(line, facility, status):
    facility_link_status = None

    if facility is not None:
        match = re.search(r"logging level {0} {1} (\S+)".format(facility, status), line, re.M)
        if match:
            facility_link_status = status + "-" + match.group(1)

    return facility_link_status


def parse_event_status(line, event):
    status = None

    match = re.search(r"logging event {0} (\S+)".format(event + "-status"), line, re.M)
    if match:
        state = match.group(1)
        if state:
            status = state

    return status


def parse_event(line):
    event = None

    match = re.search(r"logging event (\S+)", line, re.M)
    if match:
        state = match.group(1)
        if state == "link-status":
            event = "link"
        elif state == "trunk-status":
            event = "trunk"

    return event


def parse_message(line):
    message = None

    match = re.search(r"logging message interface type ethernet description", line, re.M)
    if match:
        message = "add-interface-description"

    return message


def parse_file_size(line, name, level):
    file_size = None

    match = re.search(r"logging logfile {0} {1} size (\S+)".format(name, level), line, re.M)
    if match:
        file_size = match.group(1)
        if file_size == "8192" or file_size == "4194304":
            file_size = None

    return file_size


def parse_timestamp(line):
    timestamp = None

    match = re.search(r"logging timestamp (\S+)", line, re.M)
    if match:
        timestamp = match.group(1)

    return timestamp


def parse_name(line, dest):
    name = None

    if dest is not None:
        if dest == "logfile":
            match = re.search(r"logging logfile (\S+)", line, re.M)
            if match:
                name = match.group(1)
            else:
                pass

    return name


def parse_remote_server(line, dest):
    remote_server = None

    if dest and dest == "server":
        match = re.search(r"logging server (\S+)", line, re.M)
        if match:
            remote_server = match.group(1)

    return remote_server


def parse_dest_level(line, dest, name):
    dest_level = None

    def parse_match(match):
        level = None
        if match:
            if int(match.group(1)) in range(0, 8):
                level = match.group(1)
            else:
                pass
        return level

    if dest and dest != "server":
        if dest == "logfile":
            match = re.search(r"logging logfile {0} (\S+)".format(name), line, re.M)
            if match:
                dest_level = parse_match(match)

        elif dest == "server":
            match = re.search(r"logging server (?:\S+) (\d+)", line, re.M)
            if match:
                dest_level = parse_match(match)
        else:
            match = re.search(r"logging {0} (\S+)".format(dest), line, re.M)
            if match:
                dest_level = parse_match(match)

    return dest_level


def parse_facility_level(line, facility, dest):
    facility_level = None

    if dest == "server":
        match = re.search(r"logging server (?:\S+) (\d+)", line, re.M)
        if match:
            facility_level = match.group(1)

    elif facility is not None:
        match = re.search(r"logging level {0} (\S+)".format(facility), line, re.M)
        if match:
            facility_level = match.group(1)

    return facility_level


def parse_facility(line):
    facility = None

    match = re.search(
        r"logging server (?:\S+) (?:\d+) (?:\S+) (?:\S+) (?:\S+) (\S+)",
        line,
        re.M,
    )
    if match:
        facility = match.group(1)

    return facility


def parse_use_vrf(line, dest):
    use_vrf = None

    if dest and dest == "server":
        match = re.search(r"logging server (?:\S+) (?:\d+) use-vrf (\S+)", line, re.M)
        if match:
            use_vrf = match.group(1)

    return use_vrf


def parse_interface(line):
    interface = None

    match = re.search(r"logging source-interface (\S*)", line, re.M)
    if match:
        interface = match.group(1)

    return interface


def map_config_to_obj(module):
    obj = []

    data = get_config(module, flags=[" all | section logging"])

    for line in data.split("\n"):
        if re.search(r"no (\S+)", line, re.M):
            state = "absent"
        else:
            state = "present"

        match = re.search(r"logging (\S+)", line, re.M)
        if state == "present" and match:
            event_status = None
            name = None
            dest_level = None
            dest = None
            facility = None
            remote_server = None
            facility_link_status = None
            file_size = None
            facility_level = None

            if match.group(1) in DEST_GROUP:
                dest = match.group(1)

                name = parse_name(line, dest)
                remote_server = parse_remote_server(line, dest)
                dest_level = parse_dest_level(line, dest, name)

                if dest == "server":
                    facility = parse_facility(line)

                facility_level = parse_facility_level(line, facility, dest)

                if dest == "logfile":
                    file_size = parse_file_size(line, name, dest_level)

            elif match.group(1) == "level":
                match_facility = re.search(r"logging level (\S+)", line, re.M)
                facility = match_facility.group(1)

                level = parse_facility_level(line, facility, dest)
                if level.isdigit():
                    facility_level = level
                else:
                    facility_link_status = parse_facility_link_status(line, facility, level)

            elif match.group(1) == "event" and state == "present":
                event = parse_event(line)
                if event:
                    status = parse_event_status(line, event)
                    if status:
                        event_status = event + "-" + status
                else:
                    continue

            else:
                pass

            obj.append(
                {
                    "dest": dest,
                    "remote_server": remote_server,
                    "use_vrf": parse_use_vrf(line, dest),
                    "name": name,
                    "facility": facility,
                    "dest_level": dest_level,
                    "facility_level": facility_level,
                    "interface": parse_interface(line),
                    "facility_link_status": facility_link_status,
                    "event": event_status,
                    "file_size": file_size,
                    "message": parse_message(line),
                    "timestamp": parse_timestamp(line),
                },
            )

    cmd = [
        {
            "command": "show logging | section enabled | section console",
            "output": "text",
        },
        {
            "command": "show logging | section enabled | section monitor",
            "output": "text",
        },
    ]

    default_data = run_commands(module, cmd)

    for line in default_data:
        flag = False
        match = re.search(
            r"Logging (\w+):(?:\s+) (?:\w+) (?:\W)Severity: (\w+)",
            str(line),
            re.M,
        )
        if match:
            if match.group(1) == "console" and match.group(2) == "critical":
                dest_level = "2"
                flag = True
            elif match.group(1) == "monitor" and match.group(2) == "notifications":
                dest_level = "5"
                flag = True
        if flag:
            obj.append(
                {
                    "dest": match.group(1),
                    "remote_server": None,
                    "name": None,
                    "facility": None,
                    "dest_level": dest_level,
                    "facility_level": None,
                    "use_vrf": None,
                    "interface": None,
                    "facility_link_status": None,
                    "event": None,
                    "file_size": None,
                    "message": None,
                    "timestamp": None,
                },
            )

    return obj


def map_params_to_obj(module):
    obj = []

    if "aggregate" in module.params and module.params["aggregate"]:
        args = {
            "dest": "",
            "remote_server": "",
            "use_vrf": "",
            "name": "",
            "facility": "",
            "dest_level": "",
            "facility_level": "",
            "interface": "",
            "facility_link_status": None,
            "event": None,
            "file_size": None,
            "message": None,
            "timestamp": None,
        }

        for c in module.params["aggregate"]:
            d = c.copy()

            for key in args:
                if key not in d:
                    d[key] = None

            if d["dest_level"] is not None:
                d["dest_level"] = str(d["dest_level"])

            if d["facility_level"] is not None:
                d["facility_level"] = str(d["facility_level"])

            if d["interface"]:
                d["interface"] = normalize_interface(d["interface"])

            if "state" not in d:
                d["state"] = module.params["state"]

            if d["file_size"]:
                d["file_size"] = str(d["file_size"])

            obj.append(d)

    else:
        dest_level = None
        facility_level = None
        file_size = None

        if module.params["dest_level"] is not None:
            dest_level = str(module.params["dest_level"])

        if module.params["facility_level"] is not None:
            facility_level = str(module.params["facility_level"])

        if module.params["file_size"] is not None:
            file_size = str(module.params["file_size"])

        obj.append(
            {
                "dest": module.params["dest"],
                "remote_server": module.params["remote_server"],
                "use_vrf": module.params["use_vrf"],
                "name": module.params["name"],
                "facility": module.params["facility"],
                "dest_level": dest_level,
                "facility_level": facility_level,
                "interface": normalize_interface(module.params["interface"]),
                "state": module.params["state"],
                "facility_link_status": module.params["facility_link_status"],
                "event": module.params["event"],
                "message": module.params["interface_message"],
                "file_size": file_size,
                "timestamp": module.params["timestamp"],
            },
        )
    return obj


def merge_wants(wants, want):
    if not wants:
        wants = list()

    for w in want:
        w = copy.copy(w)
        state = w["state"]
        del w["state"]

        if state == "absent":
            if w in wants:
                wants.remove(w)
        elif w not in wants:
            wants.append(w)

    return wants


def absent(h):
    h["state"] = "absent"
    return h


def outliers(haves, wants):
    wants = list(wants)
    return [absent(h) for h in haves if not (h in wants or wants.append(h))]


def main():
    """main entry point for module execution"""
    argument_spec = dict(
        dest=dict(choices=DEST_GROUP),
        name=dict(),
        facility=dict(),
        remote_server=dict(),
        use_vrf=dict(),
        dest_level=dict(type="int", aliases=["level"]),
        facility_level=dict(type="int"),
        interface=dict(),
        facility_link_status=dict(
            choices=[
                "link-down-notif",
                "link-down-error",
                "link-up-notif",
                "link-up-error",
            ],
        ),
        event=dict(
            choices=[
                "link-enable",
                "link-default",
                "trunk-enable",
                "trunk-default",
            ],
        ),
        interface_message=dict(choices=["add-interface-description"]),
        file_size=dict(type="int"),
        timestamp=dict(choices=["microseconds", "milliseconds", "seconds"]),
        state=dict(default="present", choices=["present", "absent"]),
        aggregate=dict(type="list", elements="dict"),
        purge=dict(default=False, type="bool"),
    )

    required_if = [
        ("dest", "logfile", ["name"]),
        ("dest", "server", ["remote_server"]),
    ]

    module = AnsibleModule(
        argument_spec=argument_spec,
        required_if=required_if,
        supports_check_mode=True,
    )

    warnings = list()

    result = {"changed": False}
    if warnings:
        result["warnings"] = warnings

    want = map_params_to_obj(module)
    merged_wants = merge_wants(read_module_context(module), want)
    have = map_config_to_obj(module)

    commands = map_obj_to_commands(module, (want, have))
    result["commands"] = commands

    if commands:
        if not module.check_mode:
            load_config(module, commands)
        result["changed"] = True

    save_module_context(module, merged_wants)

    if module.params.get("purge"):
        pcommands = map_obj_to_commands(module, (outliers(have, merged_wants), have))
        if pcommands:
            if not module.check_mode:
                load_config(module, pcommands)
            result["changed"] = True
        result["commands"] += pcommands

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: license_device_count_info
short_description: Information module for License Device Count
description:
- Get all License Device Count.
- Get total number of managed devices.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  device_type:
    description:
    - Device_type query parameter. Type of device.
    type: str
  registration_status:
    description:
    - Registration_status query parameter. Smart license registration status of device.
    type: str
  dna_level:
    description:
    - Dna_level query parameter. Device Cisco DNA License Level.
    type: str
  virtual_account_name:
    description:
    - Virtual_account_name query parameter. Virtual account name.
    type: str
  smart_account_id:
    description:
    - Smart_account_id query parameter. Smart account id.
    type: str
requirements:
- dnacentersdk >= 2.5.5
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Licenses DeviceCountDetails2
  description: Complete reference of the DeviceCountDetails2 API.
  link: https://developer.cisco.com/docs/dna-center/#!device-count-details-2
notes:
  - SDK Method used are
    licenses.Licenses.device_count_details2,

  - Paths used are
    get /dna/intent/api/v1/licenses/device/count,

"""

EXAMPLES = r"""
- name: Get all License Device Count
  cisco.dnac.license_device_count_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    device_type: string
    registration_status: string
    dna_level: string
    virtual_account_name: string
    smart_account_id: string
  register: result

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": 0,
      "version": "string"
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2017 Yuwei Zhou, <yuwzho@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_autoscale_info
version_added: "0.1.2"
short_description: Get Azure Auto Scale Setting facts
description:
    - Get facts of Auto Scale Setting.

options:
    resource_group:
        description:
            - The name of the resource group.
        required: True
        type: str
    name:
        description:
            - The name of the Auto Scale Setting.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Yuwei Zhou (@yuwzho)

'''

EXAMPLES = '''
- name: Get instance of Auto Scale Setting
  azure_rm_autoscale_info:
    resource_group: myResourceGroup
    name: auto_scale_name

- name: List instances of Auto Scale Setting
  azure_rm_autoscale_info:
    resource_group: myResourceGroup
    tags:
      - key
      - key:value
'''

RETURN = '''
autoscales:
    description: List of Azure Scale Settings dicts.
    returned: always
    type: list
    sample: [{
        "enabled": true,
        "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/microsoft.insights/autoscalesettings/scale",
        "location": "eastus",
        "name": "scale",
        "notifications": [
            {
                "custom_emails": [
                    "yuwzho@microsoft.com"
                ],
                "send_to_subscription_administrator": true,
                "send_to_subscription_co_administrators": false,
                "webhooks": []
            }
        ],
        "profiles": [
            {
                "count": "1",
                "max_count": "1",
                "min_count": "1",
                "name": "Auto created scale condition 0",
                "recurrence_days": [
                    "Monday"
                ],
                "recurrence_frequency": "Week",
                "recurrence_hours": [
                    "6"
                ],
                "recurrence_mins": [
                    "0"
                ],
                "recurrence_timezone": "China Standard Time",
                "rules": [
                    {
                        "cooldown": 5.0,
                        "direction": "Increase",
                        "metric_name": "Percentage CPU",
                        "metric_resource_uri": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsof
                                                t.Compute/virtualMachineScaleSets/myVmss",
                        "operator": "GreaterThan",
                        "statistic": "Average",
                        "threshold": 70.0,
                        "time_aggregation": "Average",
                        "time_grain": 1.0,
                        "time_window": 10.0,
                        "type": "ChangeCount",
                        "value": "1"
                    }
                ]
            }
        ],
        "target": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.Compute/virtualMachineScale
                   Sets/myVmss"
    }]

'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
from ansible.module_utils._text import to_native


# duplicated in azure_rm_autoscale
def timedelta_to_minutes(time):
    if not time:
        return 0
    return time.days * 1440 + time.seconds / 60.0 + time.microseconds / 60000000.0


def get_enum_value(item):
    if hasattr(item, 'value'):
        return to_native(item.value)
    return to_native(item)


def auto_scale_to_dict(instance):
    if not instance:
        return dict()
    return dict(
        id=to_native(instance.id or ''),
        name=to_native(instance.name),
        location=to_native(instance.location),
        profiles=[profile_to_dict(p) for p in instance.profiles or []],
        notifications=[notification_to_dict(n) for n in instance.notifications or []],
        enabled=instance.enabled,
        target=to_native(instance.target_resource_uri),
        tags=instance.tags
    )


def rule_to_dict(rule):
    if not rule:
        return dict()
    result = dict(metric_name=to_native(rule.metric_trigger.metric_name),
                  metric_resource_uri=to_native(rule.metric_trigger.metric_resource_uri),
                  time_grain=timedelta_to_minutes(rule.metric_trigger.time_grain),
                  statistic=get_enum_value(rule.metric_trigger.statistic),
                  time_window=timedelta_to_minutes(rule.metric_trigger.time_window),
                  time_aggregation=get_enum_value(rule.metric_trigger.time_aggregation),
                  operator=get_enum_value(rule.metric_trigger.operator),
                  threshold=float(rule.metric_trigger.threshold))
    if rule.scale_action and to_native(rule.scale_action.direction) != 'None':
        result['direction'] = get_enum_value(rule.scale_action.direction)
        result['type'] = get_enum_value(rule.scale_action.type)
        result['value'] = to_native(rule.scale_action.value)
        result['cooldown'] = timedelta_to_minutes(rule.scale_action.cooldown)
    return result


def profile_to_dict(profile):
    if not profile:
        return dict()
    result = dict(name=to_native(profile.name),
                  count=to_native(profile.capacity.default),
                  max_count=to_native(profile.capacity.maximum),
                  min_count=to_native(profile.capacity.minimum))

    if profile.rules:
        result['rules'] = [rule_to_dict(r) for r in profile.rules]
    if profile.fixed_date:
        result['fixed_date_timezone'] = profile.fixed_date.time_zone
        result['fixed_date_start'] = profile.fixed_date.start
        result['fixed_date_end'] = profile.fixed_date.end
    if profile.recurrence:
        if get_enum_value(profile.recurrence.frequency) != 'None':
            result['recurrence_frequency'] = get_enum_value(profile.recurrence.frequency)
        if profile.recurrence.schedule:
            result['recurrence_timezone'] = to_native(str(profile.recurrence.schedule.time_zone))
            result['recurrence_days'] = [to_native(r) for r in profile.recurrence.schedule.days]
            result['recurrence_hours'] = [to_native(r) for r in profile.recurrence.schedule.hours]
            result['recurrence_mins'] = [to_native(r) for r in profile.recurrence.schedule.minutes]
    return result


def notification_to_dict(notification):
    if not notification:
        return dict()
    return dict(send_to_subscription_administrator=notification.email.send_to_subscription_administrator if notification.email else False,
                send_to_subscription_co_administrators=notification.email.send_to_subscription_co_administrators if notification.email else False,
                custom_emails=[to_native(e) for e in notification.email.custom_emails or []],
                webhooks=[to_native(w.service_url) for w in notification.webhooks or []])


class AzureRMAutoScaleInfo(AzureRMModuleBase):
    def __init__(self):
        # define user inputs into argument
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            name=dict(
                type='str'
            ),
            tags=dict(
                type='list',
                elements='str'
            )
        )
        # store the results of the module operation
        self.results = dict()
        self.resource_group = None
        self.name = None
        self.tags = None

        super(AzureRMAutoScaleInfo, self).__init__(self.module_arg_spec, supports_check_mode=True, supports_tags=False, facts_module=True)

    def exec_module(self, **kwargs):

        is_old_facts = self.module._name == 'azure_rm_autoscale_facts'
        if is_old_facts:
            self.module.deprecate("The 'azure_rm_autoscale_facts' module has been renamed to 'azure_rm_autoscale_info'", version=(2.9, ))

        for key in list(self.module_arg_spec):
            setattr(self, key, kwargs[key])

        if self.resource_group and self.name:
            self.results['autoscales'] = self.get()
        elif self.resource_group:
            self.results['autoscales'] = self.list_by_resource_group()
        return self.results

    def get(self):
        result = []
        try:
            instance = self.monitor_autoscale_settings_client.autoscale_settings.get(self.resource_group, self.name)
            result = [auto_scale_to_dict(instance) if self.has_tags(instance.tags, self.tags) else None]
        except Exception as ex:
            self.log('Could not get facts for autoscale {0} - {1}.'.format(self.name, str(ex)))
        return result

    def list_by_resource_group(self):
        results = []
        try:
            response = self.monitor_autoscale_settings_client.autoscale_settings.list_by_resource_group(self.resource_group)
            results = [auto_scale_to_dict(item) for item in response if self.has_tags(item.tags, self.tags)]
        except Exception as ex:
            self.log('Could not get facts for autoscale {0} - {1}.'.format(self.name, str(ex)))
        return results


def main():
    AzureRMAutoScaleInfo()


if __name__ == '__main__':
    main()



# (c) 2022 Red Hat, Inc.
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function


__metaclass__ = type

import json
import re

from ansible.module_utils._text import to_text
from ansible.module_utils.connection import Connection


def get_connection(module):
    if hasattr(module, "_grpc_connection"):
        return module._grpc_connection

    capabilities = get_capabilities(module)
    network_api = capabilities.get("network_api")
    if network_api == "ansible.netcommon.grpc":
        module._grpc_connection = Connection(module._socket_path)
    else:
        module.fail_json(msg="Invalid connection type %s" % network_api)
    return module._grpc_connection


def get_capabilities(module):
    if hasattr(module, "_grpc_capabilities"):
        return module._grpc_capabilities

    module._grpc_capabilities = Connection(module._socket_path).get_capabilities()
    return module._grpc_capabilities


def get(module, section, data_type, check_rc=True):
    conn = get_connection(module)
    if data_type == "config":
        out = conn.get_config(section)
    else:
        out = conn.get(section)

    response = out.get("response")
    error = out.get("error")
    if error:
        if check_rc:
            module.fail_json(msg=to_text(out["error"], errors="surrogate_then_replace"))
        else:
            module.warn(to_text(out["error"], errors="surrogate_then_replace"))

    return response.strip(), error.strip()


def merge_config(module, section, check_rc=True):
    conn = get_connection(module)
    try:
        out = conn.merge_config(section)
        if out:
            err = json.loads(out)
            res = json.dumps(err, indent=4, separators=(",", ": "))
            module.fail_json(msg=to_text(res, errors="surrogate_then_replace"))
    except ValueError as err:
        if check_rc:
            module.fail_json(msg=to_text(out["error"], errors="surrogate_then_replace"))
        else:
            module.warn(to_text(out["error"], errors="surrogate_then_replace"))
        return err


def replace_config(module, section, check_rc=True):
    conn = get_connection(module)
    try:
        out = conn.replace_config(section)
        if out:
            err = json.loads(out)
            res = json.dumps(err, indent=4, separators=(",", ": "))
            module.fail_json(msg=to_text(res, errors="surrogate_then_replace"))
    except ValueError as err:
        if check_rc:
            module.fail_json(msg=to_text(out["error"], errors="surrogate_then_replace"))
        else:
            module.warn(to_text(out["error"], errors="surrogate_then_replace"))
        return err


def delete_config(module, section, check_rc=True):
    conn = get_connection(module)
    try:
        out = conn.delete_config(section)
        if out:
            err = json.loads(out)
            res = json.dumps(err, indent=4, separators=(",", ": "))
            module.fail_json(msg=to_text(res, errors="surrogate_then_replace"))
    except ValueError as err:
        if check_rc:
            module.fail_json(msg=to_text(out["error"], errors="surrogate_then_replace"))
        else:
            module.warn(to_text(out["error"], errors="surrogate_then_replace"))
        return err


def run_cli(module, command, display, check_rc=True):
    conn = get_connection(module)
    out = conn.run_cli(command, display)
    response = out.get("response")
    error = out.get("error")
    if error:
        if check_rc:
            module.fail_json(msg=to_text(out["error"], errors="surrogate_then_replace"))
        else:
            module.warn(to_text(out["error"], errors="surrogate_then_replace"))

    return response.strip(), error.strip()


def sanitize_content(data):
    out = re.sub(".*Last configuration change.*\n?", "", data)
    return out


def validate_config(module, config):
    output = ""
    params = list(config.keys())[0]
    if params:
        val = "{" + '"' + params + '": [null]}'
    response, err = get(module, val, "config")
    if response:
        output = json.loads(response)
    return output



# -*- coding: utf-8 -*-
# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is based on
# Lib/posixpath.py of cpython
#
# Copyright (c) 2001-2022 Python Software Foundation.  All rights reserved.
# It is licensed under the PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# (See LICENSES/PSF-2.0.txt in this collection)
# SPDX-License-Identifier: PSF-2.0

from __future__ import absolute_import, division, print_function

__metaclass__ = type

import os


def ismount(path):
    """Test whether a path is a mount point
    This is a copy of the upstream version of ismount(). Originally this was copied here as a workaround
    until Python issue 2466 was fixed. Now it is here so this will work on older versions of Python
    that may not have the upstream fix.
    https://github.com/ansible/ansible-modules-core/issues/2186
    http://bugs.python.org/issue2466
    """
    try:
        s1 = os.lstat(path)
    except (OSError, ValueError):
        # It doesn't exist -- so not a mount point. :-)
        return False
    else:
        # A symlink can never be a mount point
        if os.path.stat.S_ISLNK(s1.st_mode):
            return False

    if isinstance(path, bytes):
        parent = os.path.join(path, b'..')
    else:
        parent = os.path.join(path, '..')
    parent = os.path.realpath(parent)
    try:
        s2 = os.lstat(parent)
    except (OSError, ValueError):
        return False

    dev1 = s1.st_dev
    dev2 = s2.st_dev
    if dev1 != dev2:
        return True     # path/.. on a different device as path
    ino1 = s1.st_ino
    ino2 = s2.st_ino
    if ino1 == ino2:
        return True     # path/.. is the same i-node as path
    return False



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Kevin Breit (@kbreit) <kevin.breit@kevinbreit.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    "status": ['deprecated'],
    'supported_by': 'community'
}

DOCUMENTATION = r'''
---
module: meraki_mr_settings
short_description: Manage general settings for Meraki wireless networks
description:
- Allows for configuration of general settings in Meraki MR wireless networks.
deprecated:
  removed_in: '3.0.0'
  why: Updated modules released with increased functionality
  alternative: cisco.meraki.networks_wireless_settings
options:
    state:
        description:
        - Query or edit wireless settings.
        type: str
        choices: [ present, query]
        default: present
    net_name:
        description:
        - Name of network.
        type: str
    net_id:
        description:
        - ID of network.
        type: str
    upgrade_strategy:
        description:
        - The upgrade strategy to apply to the network.
        - Requires firmware version MR 26.8 or higher.
        choices: [ minimize_upgrade_time, minimize_client_downtime ]
        type: str
    ipv6_bridge_enabled:
        description:
        - Toggle for enabling or disabling IPv6 bridging in a network.
        - If enabled, SSIDs must also be configured to use bridge mode.
        type: bool
    led_lights_on:
        description:
        - Toggle for enabling or disabling LED lights on all APs in the network.
        type: bool
    location_analytics_enabled:
        description:
        - Toggle for enabling or disabling location analytics for your network.
        type: bool
    meshing_enabled:
        description: Toggle for enabling or disabling meshing in a network.
        type: bool
author:
- Kevin Breit (@kbreit)
extends_documentation_fragment: cisco.meraki.meraki
'''

EXAMPLES = r'''
- name: Query all settings
  meraki_mr_settings:
    auth_key: abc123
    org_name: YourOrg
    net_name: YourNet
    state: query
  delegate_to: localhost
- name: Configure settings
  meraki_mr_settings:
    auth_key: abc123
    org_name: YourOrg
    net_name: YourNet
    state: present
    upgrade_strategy: minimize_upgrade_time
    ipv6_bridge_enabled: false
    led_lights_on: true
    location_analytics_enabled: true
    meshing_enabled: true
  delegate_to: localhost
'''

RETURN = r'''
data:
    description: List of wireless settings.
    returned: success
    type: complex
    contains:
        upgrade_strategy:
            description:
            - The upgrade strategy to apply to the network.
            - Requires firmware version MR 26.8 or higher.
            type: str
            returned: success
            sample: minimize_upgrade_time
        ipv6_bridge_enabled:
            description:
            - Toggle for enabling or disabling IPv6 bridging in a network.
            - If enabled, SSIDs must also be configured to use bridge mode.
            type: bool
            returned: success
            sample: True
        led_lights_on:
            description:
            - Toggle for enabling or disabling LED lights on all APs in the network.
            type: bool
            returned: success
            sample: True
        location_analytics_enabled:
            description:
            - Toggle for enabling or disabling location analytics for your network.
            type: bool
            returned: success
            sample: True
        meshing_enabled:
            description: Toggle for enabling or disabling meshing in a network.
            type: bool
            returned: success
            sample: True
'''

from ansible.module_utils.basic import AnsibleModule, json
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import MerakiModule, meraki_argument_spec
from re import sub


def convert_to_camel_case(string):
    string = sub(r"(_|-)+", " ", string).title().replace(" ", "")
    return string[0].lower() + string[1:]


def construct_payload(meraki):
    payload = {}
    if meraki.params['upgrade_strategy'] is not None:
        payload['upgradeStrategy'] = convert_to_camel_case(meraki.params['upgrade_strategy'])
    if meraki.params['ipv6_bridge_enabled'] is not None:
        payload['ipv6BridgeEnabled'] = meraki.params['ipv6_bridge_enabled']
    if meraki.params['led_lights_on'] is not None:
        payload['ledLightsOn'] = meraki.params['led_lights_on']
    if meraki.params['location_analytics_enabled'] is not None:
        payload['locationAnalyticsEnabled'] = meraki.params['location_analytics_enabled']
    if meraki.params['meshing_enabled'] is not None:
        payload['meshingEnabled'] = meraki.params['meshing_enabled']
    return payload


def main():
    # define the available arguments/parameters that a user can pass to
    # the module
    argument_spec = meraki_argument_spec()
    argument_spec.update(state=dict(type='str', choices=['present', 'query'], default='present'),
                         org_name=dict(type='str', aliases=['organization']),
                         org_id=dict(type='str'),
                         net_name=dict(type='str'),
                         net_id=dict(type='str'),
                         upgrade_strategy=dict(type='str', choices=['minimize_upgrade_time',
                                                                    'minimize_client_downtime']),
                         ipv6_bridge_enabled=dict(type='bool'),
                         led_lights_on=dict(type='bool'),
                         location_analytics_enabled=dict(type='bool'),
                         meshing_enabled=dict(type='bool'),
                         )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True,
                           )
    meraki = MerakiModule(module, function='mr_settings')

    meraki.params['follow_redirects'] = 'all'

    query_urls = {'mr_settings': '/networks/{net_id}/wireless/settings'}
    update_urls = {'mr_settings': '/networks/{net_id}/wireless/settings'}

    meraki.url_catalog['get_one'].update(query_urls)
    meraki.url_catalog['update'] = update_urls

    org_id = meraki.params['org_id']
    net_id = meraki.params['net_id']
    if org_id is None:
        org_id = meraki.get_org_id(meraki.params['org_name'])
    if net_id is None:
        nets = meraki.get_nets(org_id=org_id)
        net_id = meraki.get_net_id(org_id, meraki.params['net_name'], data=nets)

    if meraki.params['state'] == 'query':
        path = meraki.construct_path('get_one', net_id=net_id)
        response = meraki.request(path, method='GET')
        meraki.result['data'] = response
        meraki.exit_json(**meraki.result)
    elif meraki.params['state'] == 'present':
        path = meraki.construct_path('get_one', net_id=net_id)
        original = meraki.request(path, method='GET')
        payload = construct_payload(meraki)
        if meraki.is_update_required(original, payload) is True:
            if meraki.check_mode is True:
                meraki.result['data'] = payload
                meraki.result['changed'] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path('update', net_id=net_id)
            response = meraki.request(path, method='PUT', payload=json.dumps(payload))
            meraki.result['data'] = response
            meraki.result['changed'] = True
            meraki.exit_json(**meraki.result)
        else:
            meraki.result['data'] = original
            meraki.exit_json(**meraki.result)

    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == '__main__':
    main()



# Copyright (c) 2019-2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


from ansible.compat import selectors

from ansible_collections.community.docker.plugins.module_utils.socket_handler import (
    DockerSocketHandlerBase,
)


class DockerSocketHandler(DockerSocketHandlerBase):
    def __init__(self, display, sock, log=None, container=None):
        super(DockerSocketHandler, self).__init__(sock, selectors, log=lambda msg: display.vvvv(msg, host=container))



#
# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

import json
import re

from ansible.errors import AnsibleConnectionFailure
from ansible.module_utils._text import to_bytes, to_text
from ansible_collections.ansible.netcommon.plugins.plugin_utils.terminal_base import TerminalBase


class TerminalModule(TerminalBase):
    terminal_stdout_re = [
        re.compile(rb"[\r\n]?[\w+\-\.:\/\[\]]+(?:\([^\)]+\)){,3}(?:>|#) ?$"),
        re.compile(rb"\[\w+\@[\w\-\.]+(?: [^\]])\] ?[>#\$] ?$"),
    ]

    terminal_stderr_re = [
        re.compile(rb"% ?Error"),
        # re.compile(br"^% \w+", re.M),
        re.compile(rb"% User not present"),
        re.compile(rb"% ?Bad secret"),
        re.compile(rb"invalid input", re.I),
        re.compile(rb"(?:incomplete|ambiguous) command", re.I),
        re.compile(rb"connection timed out", re.I),
        # Strings like this regarding VLANs are not errors
        re.compile(rb"[^\r\n]+ not found(?! in current VLAN)", re.I),
        re.compile(rb"'[^']' +returned error code: ?\d+"),
        re.compile(rb"[^\r\n](?<! shell )\/bin\/(?:ba)?sh"),
        re.compile(rb"% More than \d+ OSPF instance", re.I),
        re.compile(rb"% Subnet [0-9a-f.:/]+ overlaps", re.I),
        re.compile(rb"Maximum number of pending sessions has been reached"),
        re.compile(rb"% Prefix length must be less than"),
        # returned in response to 'channel-group <name> mode <mode>'
        re.compile(
            rb"% Cannot change mode; remove all members and try again.",
        ),
    ]

    terminal_config_prompt = re.compile(r"^.+\(config(-.*)?\)#$")

    def on_open_shell(self):
        try:
            for cmd in (b"terminal length 0", b"terminal width 512"):
                self._exec_cli_command(cmd)
        except AnsibleConnectionFailure:
            raise AnsibleConnectionFailure("unable to set terminal parameters")

    def on_become(self, passwd=None):
        if self._get_prompt().endswith(b"#"):
            return

        cmd = {"command": "enable"}
        if passwd:
            cmd["prompt"] = to_text(
                r"[\r\n]?[Pp]assword: $",
                errors="surrogate_or_strict",
            )
            cmd["answer"] = passwd
            cmd["prompt_retry_check"] = True

        try:
            self._exec_cli_command(
                to_bytes(json.dumps(cmd), errors="surrogate_or_strict"),
            )
            prompt = self._get_prompt()
            if prompt is None or not prompt.endswith(b"#"):
                raise AnsibleConnectionFailure(
                    "failed to elevate privilege to enable mode still at prompt [%s]" % prompt,
                )
        except AnsibleConnectionFailure as e:
            prompt = self._get_prompt()
            raise AnsibleConnectionFailure(
                "unable to elevate privilege to enable mode, at prompt [%s] with error: %s"
                % (prompt, e.message),
            )

    def on_unbecome(self):
        prompt = self._get_prompt()
        if prompt is None:
            # if prompt is None most likely the terminal is hung up at a prompt
            return

        if b"(config" in prompt:
            self._exec_cli_command(b"end")
            self._exec_cli_command(b"disable")

        elif prompt.endswith(b"#"):
            self._exec_cli_command(b"disable")



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The eos_acl_interfaces class
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to it's desired end-state is
created
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

import itertools

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base import (
    ConfigBase,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    search_obj_in_list,
    to_list,
)

from ansible_collections.arista.eos.plugins.module_utils.network.eos.facts.facts import Facts


class Acl_interfaces(ConfigBase):
    """
    The eos_acl_interfaces class
    """

    gather_subset = ["!all", "!min"]

    gather_network_resources = ["acl_interfaces"]

    def __init__(self, module):
        super(Acl_interfaces, self).__init__(module)

    def get_acl_interfaces_facts(self, data=None):
        """Get the 'facts' (the current configuration)

        :rtype: A dictionary
        :returns: The current configuration as a dictionary
        """
        facts, _warnings = Facts(self._module).get_facts(
            self.gather_subset,
            self.gather_network_resources,
            data=data,
        )
        acl_interfaces_facts = facts["ansible_network_resources"].get(
            "acl_interfaces",
        )
        if not acl_interfaces_facts:
            return []
        return acl_interfaces_facts

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        result = {"changed": False}
        warnings = list()
        commands = list()
        changed = False

        if self.state in self.ACTION_STATES:
            existing_acl_interfaces_facts = self.get_acl_interfaces_facts()
        else:
            existing_acl_interfaces_facts = []
        if self.state in self.ACTION_STATES or self.state == "rendered":
            commands.extend(self.set_config(existing_acl_interfaces_facts))
        if commands and self.state in self.ACTION_STATES:
            if not self._module.check_mode:
                self._connection.edit_config(commands)
                changed = True
            if changed:
                result["changed"] = True
        if self.state in self.ACTION_STATES:
            result["commands"] = commands
        if self.state in self.ACTION_STATES or self.state == "gathered":
            changed_acl_interfaces_facts = self.get_acl_interfaces_facts()
        elif self.state == "rendered":
            result["rendered"] = commands
        elif self.state == "parsed":
            if not self._module.params["running_config"]:
                self._module.fail_json(
                    msg="Value of running_config parameter must not be empty for state parsed",
                )
            result["parsed"] = self.get_acl_interfaces_facts(
                data=self._module.params["running_config"],
            )
        else:
            changed_acl_interfaces_facts = []
        if self.state in self.ACTION_STATES:
            result["before"] = existing_acl_interfaces_facts
            if result["changed"]:
                result["after"] = changed_acl_interfaces_facts
        elif self.state == "gathered":
            result["gathered"] = changed_acl_interfaces_facts

        result["warnings"] = warnings
        return result

    def set_config(self, existing_acl_interfaces_facts):
        """Collect the configuration from the args passed to the module,
            collect the current configuration (as a dict from facts)

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        want = self._module.params["config"]
        have = existing_acl_interfaces_facts
        resp = self.set_state(want, have)
        return to_list(resp)

    def set_state(self, want, have):
        """Select the appropriate function based on the state provided

        :param want: the desired configuration as a dictionary
        :param have: the current configuration as a dictionary
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []
        if self.state in ("merged", "replaced", "overridden") and not want:
            self._module.fail_json(
                msg="value of config parameter must not be empty for state {0}".format(
                    self.state,
                ),
            )
        state = self._module.params["state"]
        if state == "overridden":
            commands = self._state_overridden(want, have)
        elif state == "deleted":
            commands = self._state_deleted(want, have)
        elif state == "merged" or self.state == "rendered":
            commands = self._state_merged(want, have)
        elif state == "replaced":
            commands = self._state_replaced(want, have)
        return commands

    def _state_replaced(self, want, have):
        """The command generator when state is replaced

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commandset = []
        want_interface = []
        for w in want:
            commands = []
            diff_access_group = []
            want_interface.append(w["name"])
            obj_in_have = search_obj_in_list(w["name"], have, "name")
            if not obj_in_have or "access_groups" not in obj_in_have.keys():
                commands.append(add_commands(w["access_groups"], w["name"]))
            else:
                if "access_groups" in obj_in_have.keys():
                    obj = self.get_acl_diff(obj_in_have, w)
                    if obj[0]:
                        to_delete = {
                            "access_groups": [{"acls": obj[0], "afi": "ipv4"}],
                        }
                        commands.append(remove_commands(to_delete, w["name"]))
                    if obj[1]:
                        to_delete = {
                            "access_groups": [{"acls": obj[1], "afi": "ipv6"}],
                        }
                        commands.append(remove_commands(to_delete, w["name"]))
                    diff = self.get_acl_diff(w, obj_in_have)
                    if diff[0]:
                        diff_access_group.append(
                            {"afi": "ipv4", "acls": diff[0]},
                        )
                    if diff[1]:
                        diff_access_group.append(
                            {"afi": "ipv6", "acls": diff[1]},
                        )
                    if diff_access_group:
                        commands.append(
                            add_commands(diff_access_group, w["name"]),
                        )
            if commands:
                intf_command = ["interface " + w["name"]]
                commands = list(itertools.chain(*commands))
                commandset.append(intf_command)
                commandset.append(commands)

        if commandset:
            commandset = list(itertools.chain(*commandset))
        return commandset

    def _state_overridden(self, want, have):
        """The command generator when state is overridden

        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commandset = []
        want_interface = []
        for w in want:
            commands = []
            diff_access_group = []
            want_interface.append(w["name"])
            obj_in_have = search_obj_in_list(w["name"], have, "name")
            if not obj_in_have or "access_groups" not in obj_in_have.keys():
                commands.append(add_commands(w["access_groups"], w["name"]))
            else:
                if "access_groups" in obj_in_have.keys():
                    obj = self.get_acl_diff(obj_in_have, w)
                    if obj[0]:
                        to_delete = {
                            "access_groups": [{"acls": obj[0], "afi": "ipv4"}],
                        }
                        commands.append(remove_commands(to_delete, w["name"]))
                    if obj[1]:
                        to_delete = {
                            "access_groups": [{"acls": obj[1], "afi": "ipv6"}],
                        }
                        commands.append(remove_commands(to_delete, w["name"]))
                    diff = self.get_acl_diff(w, obj_in_have)
                    if diff[0]:
                        diff_access_group.append(
                            {"afi": "ipv4", "acls": diff[0]},
                        )
                    if diff[1]:
                        diff_access_group.append(
                            {"afi": "ipv6", "acls": diff[1]},
                        )
                    if diff_access_group:
                        commands.append(
                            add_commands(diff_access_group, w["name"]),
                        )
            if commands:
                intf_command = ["interface " + w["name"]]
                commands = list(itertools.chain(*commands))
                commandset.append(intf_command)
                commandset.append(commands)
        for h in have:
            commands = []
            if "access_groups" in h.keys() and h["access_groups"]:
                if h["name"] not in want_interface:
                    for h_group in h["access_groups"]:
                        to_delete = {"access_groups": [h_group]}
                        commands.append(remove_commands(to_delete, h["name"]))
            if commands:
                intf_command = ["interface " + h["name"]]
                commands = list(itertools.chain(*commands))
                commandset.append(intf_command)
                commandset.append(commands)

        if commandset:
            commandset = list(itertools.chain(*commandset))

        return commandset

    def _state_merged(self, want, have):
        """The command generator when state is merged

        :rtype: A list
        :returns: the commands necessary to merge the provided into
                  the current configuration
        """
        commandset = []
        for w in want:
            commands = []
            diff_access_group = []
            obj_in_have = search_obj_in_list(w["name"], have, "name")
            if not obj_in_have:
                commands = add_commands(w["access_groups"], w["name"])
            else:
                if "access_groups" in obj_in_have.keys():
                    diff = self.get_acl_diff(w, obj_in_have)
                    if diff[0]:
                        diff_access_group.append(
                            {"afi": "ipv4", "acls": diff[0]},
                        )
                    if diff[1]:
                        diff_access_group.append(
                            {"afi": "ipv6", "acls": diff[1]},
                        )
                    if diff_access_group:
                        commands = add_commands(diff_access_group, w["name"])
                else:
                    commands = add_commands(w["access_groups"], w["name"])
            if commands:
                intf_command = ["interface " + w["name"]]
                commandset.append(intf_command)
                commandset.append(commands)
        if commandset:
            commandset = list(itertools.chain(*commandset))
        return commandset

    def _state_deleted(self, want, have):
        """The command generator when state is deleted

        :rtype: A list
        :returns: the commands necessary to remove the current configuration
                  of the provided objects
        """
        commandset = []
        for w in want:
            commands = []
            intf_command = ["interface " + w["name"]]
            obj_in_have = search_obj_in_list(w["name"], have, "name")
            if "access_groups" not in w.keys() or not w["access_groups"]:
                commands = remove_commands(obj_in_have, w["name"])
            if w["access_groups"]:
                for w_grp in w["access_groups"]:
                    if "acls" not in w_grp.keys() or not w_grp["acls"]:
                        obj = self.get_acls_from_afi(
                            w["name"],
                            w_grp["afi"],
                            have,
                        )
                        to_delete = {
                            "access_groups": [
                                {"acls": obj, "afi": w_grp["afi"]},
                            ],
                        }
                        commands = remove_commands(to_delete, w["name"])
                    else:
                        if (
                            "access_groups" not in obj_in_have.keys()
                            or not obj_in_have["access_groups"]
                        ):
                            continue
                        group = {"access_groups": [w_grp]}
                        obj = self.get_acl_diff(group, obj_in_have, True)
                        if obj[0]:
                            to_delete = {
                                "access_groups": [
                                    {"acls": obj[0], "afi": "ipv4"},
                                ],
                            }
                            commands.append(
                                remove_commands(to_delete, w["name"]),
                            )
                        if obj[1]:
                            to_delete = {
                                "access_groups": [
                                    {"acls": obj[1], "afi": "ipv6"},
                                ],
                            }
                            commands.append(
                                remove_commands(to_delete, w["name"]),
                            )
                        if commands:
                            commands = list(itertools.chain(*commands))
            if commands:
                commandset.append(intf_command)
                commandset.append(commands)

        if commandset:
            commandset = list(itertools.chain(*commandset))
        return commandset

    def get_acl_diff(self, w, h, intersection=False):
        diff_v4 = []
        diff_v6 = []
        w_acls_v4 = []
        w_acls_v6 = []
        h_acls_v4 = []
        h_acls_v6 = []
        for w_group in w["access_groups"]:
            if w_group["afi"] == "ipv4":
                w_acls_v4 = w_group["acls"]
            if w_group["afi"] == "ipv6":
                w_acls_v6 = w_group["acls"]
        for h_group in h["access_groups"]:
            if h_group["afi"] == "ipv4":
                h_acls_v4 = h_group["acls"]
            if h_group["afi"] == "ipv6":
                h_acls_v6 = h_group["acls"]
        for item in w_acls_v4:
            match = list(
                filter(lambda x: x["name"] == item["name"], h_acls_v4),
            )
            if match:
                if item["direction"] == match[0]["direction"]:
                    if intersection:
                        diff_v4.append(item)
                else:
                    if not intersection:
                        diff_v4.append(item)
            else:
                if not intersection:
                    diff_v4.append(item)
        for item in w_acls_v6:
            match = list(
                filter(lambda x: x["name"] == item["name"], h_acls_v6),
            )
            if match:
                if item["direction"] == match[0]["direction"]:
                    if intersection:
                        diff_v6.append(item)
                else:
                    if not intersection:
                        diff_v6.append(item)
            else:
                if not intersection:
                    diff_v6.append(item)
        return diff_v4, diff_v6

    def get_acls_from_afi(self, interface, afi, have):
        config = []
        for h in have:
            if h["name"] == interface:
                if "access_groups" not in h.keys() or not h["access_groups"]:
                    continue
                if h["access_groups"]:
                    for h_grp in h["access_groups"]:
                        if h_grp["afi"] == afi:
                            config = h_grp["acls"]
        return config


def add_commands(want, interface):
    commands = []

    for w in want:
        # This module was verified on an ios device since vEOS doesnot support
        # acl_interfaces cnfiguration. In ios, ipv6 acl is configured as
        # traffic-filter and in eos it is access-group

        # a_cmd = "traffic-filter" if w['afi'] == 'ipv6' else "access-group"
        a_cmd = "access-group"
        afi = "ip" if w["afi"] == "ipv4" else w["afi"]
        if "acls" in w.keys():
            for acl in w["acls"]:
                commands.append(
                    afi + " " + a_cmd + " " + acl["name"] + " " + acl["direction"],
                )
    return commands


def remove_commands(want, interface):
    commands = []
    if "access_groups" not in want.keys() or not want["access_groups"]:
        return commands
    for w in want["access_groups"]:
        # This module was verified on an ios device since vEOS doesnot support
        # acl_interfaces cnfiguration. In ios, ipv6 acl is configured as
        # traffic-filter and in eos it is access-group

        # a_cmd = "traffic-filter" if w['afi'] == 'ipv6' else "access-group"
        a_cmd = "access-group"

        afi = "ip" if w["afi"] == "ipv4" else w["afi"]
        if "acls" in w.keys():
            for acl in w["acls"]:
                commands.append(
                    "no " + afi + " " + a_cmd + " " + acl["name"] + " " + acl["direction"],
                )
    return commands



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# This module_utils is PRIVATE and should only be used by this collection. Breaking changes can occur any time.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


import traceback

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.dns.plugins.module_utils.argspec import (
    ArgumentSpec,
)

from ansible_collections.community.dns.plugins.module_utils.zone_record_api import (
    DNSAPIError,
    DNSAPIAuthenticationError,
)

from ._utils import (
    normalize_dns_name,
)


def create_module_argument_spec(provider_information):
    return ArgumentSpec(
        argument_spec=dict(
            zone_name=dict(type='str', aliases=['zone']),
            zone_id=dict(type=provider_information.get_zone_id_type()),
        ),
        required_one_of=[
            ('zone_name', 'zone_id'),
        ],
        mutually_exclusive=[
            ('zone_name', 'zone_id'),
        ],
    )


def run_module(module, create_api, provider_information):
    try:
        # Create API
        api = create_api()

        # Get zone information
        if module.params.get('zone_name') is not None:
            zone_id = normalize_dns_name(module.params.get('zone_name'))
            zone = api.get_zone_by_name(zone_id)
            if zone is None:
                module.fail_json(msg='Zone not found')
        else:
            zone = api.get_zone_by_id(module.params.get('zone_id'))
            if zone is None:
                module.fail_json(msg='Zone not found')

        module.exit_json(
            changed=False,
            zone_name=zone.name,
            zone_id=zone.id,
            zone_info=zone.info,
        )
    except DNSAPIAuthenticationError as e:
        module.fail_json(msg='Cannot authenticate: {0}'.format(e), error=to_text(e), exception=traceback.format_exc())
    except DNSAPIError as e:
        module.fail_json(msg='Error: {0}'.format(e), error=to_text(e), exception=traceback.format_exc())



# -*- coding: utf-8 -*-
# Copyright (c) 2023, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
    name: onepassword_doc
    author:
      - Sam Doran (@samdoran)
    requirements:
      - C(op) 1Password command line utility version 2 or later.
    short_description: Fetch documents stored in 1Password
    version_added: "8.1.0"
    description:
      - P(community.general.onepassword_doc#lookup) wraps C(op) command line utility to fetch one or more documents from 1Password.
    notes:
      - The document contents are a string exactly as stored in 1Password.
      - This plugin requires C(op) version 2 or later.

    options:
      _terms:
        description: Identifier(s) (case-insensitive UUID or name) of item(s) to retrieve.
        required: true

    extends_documentation_fragment:
      - community.general.onepassword
      - community.general.onepassword.lookup
'''

EXAMPLES = """
- name: Retrieve a private key from 1Password
  ansible.builtin.debug:
    var: lookup('community.general.onepassword_doc', 'Private key')
"""

RETURN = """
  _raw:
    description: Requested document
    type: list
    elements: string
"""

from ansible_collections.community.general.plugins.lookup.onepassword import OnePass, OnePassCLIv2
from ansible.errors import AnsibleLookupError
from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins.lookup import LookupBase


class OnePassCLIv2Doc(OnePassCLIv2):
    def get_raw(self, item_id, vault=None, token=None):
        args = ["document", "get", item_id]
        if vault is not None:
            args = [*args, "--vault={0}".format(vault)]

        if self.service_account_token:
            if vault is None:
                raise AnsibleLookupError("'vault' is required with 'service_account_token'")

            environment_update = {"OP_SERVICE_ACCOUNT_TOKEN": self.service_account_token}
            return self._run(args, environment_update=environment_update)

        if token is not None:
            args = [*args, to_bytes("--session=") + token]

        return self._run(args)


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        self.set_options(var_options=variables, direct=kwargs)

        vault = self.get_option("vault")
        subdomain = self.get_option("subdomain")
        domain = self.get_option("domain", "1password.com")
        username = self.get_option("username")
        secret_key = self.get_option("secret_key")
        master_password = self.get_option("master_password")
        service_account_token = self.get_option("service_account_token")
        account_id = self.get_option("account_id")
        connect_host = self.get_option("connect_host")
        connect_token = self.get_option("connect_token")

        op = OnePass(
            subdomain=subdomain,
            domain=domain,
            username=username,
            secret_key=secret_key,
            master_password=master_password,
            service_account_token=service_account_token,
            account_id=account_id,
            connect_host=connect_host,
            connect_token=connect_token,
            cli_class=OnePassCLIv2Doc,
        )
        op.assert_logged_in()

        values = []
        for term in terms:
            values.append(op.get_raw(term, vault))

        return values



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: csr_delete
short_description: Resource module for CSR Delete
description:
- Manage operation delete of the resource CSR Delete.
- This API deletes the certificate signing request of a particular node based on a given hostname and ID.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.ise.module
author: Rafael Campos (@racampos)
options:
  hostName:
    description: HostName path parameter. Name of the host of which CSR's should be
      deleted.
    type: str
  id:
    description: Id path parameter. ID of the Certificate Signing Request to be deleted.
    type: str
requirements:
- ciscoisesdk >= 2.1.1
- python >= 3.5
seealso:
- name: Cisco ISE documentation for Certificates
  description: Complete reference of the Certificates API.
  link: https://developer.cisco.com/docs/identity-services-engine/v1/#!certificate-openapi
notes:
  - SDK Method used are
    certificates.Certificates.delete_csr_by_id,

  - Paths used are
    delete /api/v1/certs/certificate-signing-request/{hostName}/{id},

"""

EXAMPLES = r"""
- name: Delete by id
  cisco.ise.csr_delete:
    ise_hostname: "{{ise_hostname}}"
    ise_username: "{{ise_username}}"
    ise_password: "{{ise_password}}"
    ise_verify: "{{ise_verify}}"
    hostName: string
    id: string

"""

RETURN = r"""
ise_response:
  description: A dictionary or list with the response returned by the Cisco ISE Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "message": "string"
      },
      "version": "string"
    }
"""



# -*- coding: utf-8 -*-

# Copyright (c) 2017 Ansible Project
#
# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module adds shared support for Direct Connect modules.
"""

import traceback

try:
    import botocore
except ImportError:
    pass

from .retries import AWSRetry


class DirectConnectError(Exception):
    def __init__(self, msg, last_traceback=None, exception=None):
        self.msg = msg
        self.last_traceback = last_traceback
        self.exception = exception


def delete_connection(client, connection_id):
    try:
        AWSRetry.jittered_backoff()(client.delete_connection)(connectionId=connection_id)
    except botocore.exceptions.ClientError as e:
        raise DirectConnectError(
            msg=f"Failed to delete DirectConnection {connection_id}.",
            last_traceback=traceback.format_exc(),
            exception=e,
        )


def associate_connection_and_lag(client, connection_id, lag_id):
    try:
        AWSRetry.jittered_backoff()(client.associate_connection_with_lag)(connectionId=connection_id, lagId=lag_id)
    except botocore.exceptions.ClientError as e:
        raise DirectConnectError(
            msg=f"Failed to associate Direct Connect connection {connection_id} with link aggregation group {lag_id}.",
            last_traceback=traceback.format_exc(),
            exception=e,
        )


def disassociate_connection_and_lag(client, connection_id, lag_id):
    try:
        AWSRetry.jittered_backoff()(client.disassociate_connection_from_lag)(connectionId=connection_id, lagId=lag_id)
    except botocore.exceptions.ClientError as e:
        raise DirectConnectError(
            msg=f"Failed to disassociate Direct Connect connection {connection_id} from link aggregation group {lag_id}.",
            last_traceback=traceback.format_exc(),
            exception=e,
        )


def delete_virtual_interface(client, virtual_interface):
    try:
        AWSRetry.jittered_backoff()(client.delete_virtual_interface)(virtualInterfaceId=virtual_interface)
    except botocore.exceptions.ClientError as e:
        raise DirectConnectError(
            msg=f"Could not delete virtual interface {virtual_interface}",
            last_traceback=traceback.format_exc(),
            exception=e,
        )



# -*- coding: utf-8 -*-
#
# Copyright (c) 2019, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function
__metaclass__ = type


import base64
import binascii
import re
import sys
import traceback

from ansible.module_utils.common.text.converters import to_text, to_bytes, to_native
from ansible.module_utils.six.moves.urllib.parse import urlparse, urlunparse, ParseResult

from ._asn1 import serialize_asn1_string_as_der

from ansible_collections.community.crypto.plugins.module_utils.version import LooseVersion

try:
    import cryptography
    from cryptography import x509
    from cryptography.exceptions import InvalidSignature
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import padding
    import ipaddress
except ImportError:
    # Error handled in the calling module.
    pass

try:
    import cryptography.hazmat.primitives.asymmetric.rsa
except ImportError:
    pass
try:
    import cryptography.hazmat.primitives.asymmetric.ec
except ImportError:
    pass
try:
    import cryptography.hazmat.primitives.asymmetric.dsa
except ImportError:
    pass
try:
    import cryptography.hazmat.primitives.asymmetric.ed25519
except ImportError:
    pass
try:
    import cryptography.hazmat.primitives.asymmetric.ed448
except ImportError:
    pass

try:
    # This is a separate try/except since this is only present in cryptography 2.5 or newer
    from cryptography.hazmat.primitives.serialization.pkcs12 import (
        load_key_and_certificates as _load_key_and_certificates,
    )
except ImportError:
    # Error handled in the calling module.
    _load_key_and_certificates = None

try:
    # This is a separate try/except since this is only present in cryptography 36.0.0 or newer
    from cryptography.hazmat.primitives.serialization.pkcs12 import (
        load_pkcs12 as _load_pkcs12,
    )
except ImportError:
    # Error handled in the calling module.
    _load_pkcs12 = None

try:
    import idna

    HAS_IDNA = True
except ImportError:
    HAS_IDNA = False
    IDNA_IMP_ERROR = traceback.format_exc()

from ansible.module_utils.basic import missing_required_lib

from .basic import (
    CRYPTOGRAPHY_HAS_DSA_SIGN,
    CRYPTOGRAPHY_HAS_EC_SIGN,
    CRYPTOGRAPHY_HAS_ED25519,
    CRYPTOGRAPHY_HAS_ED25519_SIGN,
    CRYPTOGRAPHY_HAS_ED448,
    CRYPTOGRAPHY_HAS_ED448_SIGN,
    CRYPTOGRAPHY_HAS_RSA_SIGN,
    CRYPTOGRAPHY_HAS_X25519,
    CRYPTOGRAPHY_HAS_X25519_FULL,
    CRYPTOGRAPHY_HAS_X448,
    OpenSSLObjectError,
)

from ._objects import (
    OID_LOOKUP,
    OID_MAP,
    NORMALIZE_NAMES_SHORT,
    NORMALIZE_NAMES,
)

from ._obj2txt import obj2txt


DOTTED_OID = re.compile(r'^\d+(?:\.\d+)+$')


def cryptography_get_extensions_from_cert(cert):
    result = dict()
    try:
        # Since cryptography will not give us the DER value for an extension
        # (that is only stored for unrecognized extensions), we have to re-do
        # the extension parsing ourselves.
        backend = default_backend()
        try:
            # For certain old versions of cryptography, backend is a MultiBackend object,
            # which has no _lib attribute. In that case, revert to the old approach.
            backend._lib
        except AttributeError:
            backend = cert._backend

        x509_obj = cert._x509
        # With cryptography 35.0.0, we can no longer use obj2txt. Unfortunately it still does
        # not allow to get the raw value of an extension, so we have to use this ugly hack:
        exts = list(cert.extensions)

        for i in range(backend._lib.X509_get_ext_count(x509_obj)):
            ext = backend._lib.X509_get_ext(x509_obj, i)
            if ext == backend._ffi.NULL:
                continue
            crit = backend._lib.X509_EXTENSION_get_critical(ext)
            data = backend._lib.X509_EXTENSION_get_data(ext)
            backend.openssl_assert(data != backend._ffi.NULL)
            der = backend._ffi.buffer(data.data, data.length)[:]
            entry = dict(
                critical=(crit == 1),
                value=to_native(base64.b64encode(der)),
            )
            try:
                oid = obj2txt(backend._lib, backend._ffi, backend._lib.X509_EXTENSION_get_object(ext))
            except AttributeError:
                oid = exts[i].oid.dotted_string
            result[oid] = entry

    except Exception:
        # In case the above method breaks, we likely have cryptography 36.0.0 or newer.
        # Use it's public_bytes() feature in that case. We will later switch this around
        # so that this code will be the default, but for now this will act as a fallback
        # since it will re-serialize de-serialized data, which can be different (if the
        # original data was not canonicalized) from what was contained in the certificate.
        for ext in cert.extensions:
            result[ext.oid.dotted_string] = dict(
                critical=ext.critical,
                value=to_native(base64.b64encode(ext.value.public_bytes())),
            )

    return result


def cryptography_get_extensions_from_csr(csr):
    result = dict()
    try:
        # Since cryptography will not give us the DER value for an extension
        # (that is only stored for unrecognized extensions), we have to re-do
        # the extension parsing ourselves.
        backend = default_backend()
        try:
            # For certain old versions of cryptography, backend is a MultiBackend object,
            # which has no _lib attribute. In that case, revert to the old approach.
            backend._lib
        except AttributeError:
            backend = csr._backend

        extensions = backend._lib.X509_REQ_get_extensions(csr._x509_req)
        extensions = backend._ffi.gc(
            extensions,
            lambda ext: backend._lib.sk_X509_EXTENSION_pop_free(
                ext,
                backend._ffi.addressof(backend._lib._original_lib, "X509_EXTENSION_free")
            )
        )

        # With cryptography 35.0.0, we can no longer use obj2txt. Unfortunately it still does
        # not allow to get the raw value of an extension, so we have to use this ugly hack:
        exts = list(csr.extensions)

        for i in range(backend._lib.sk_X509_EXTENSION_num(extensions)):
            ext = backend._lib.sk_X509_EXTENSION_value(extensions, i)
            if ext == backend._ffi.NULL:
                continue
            crit = backend._lib.X509_EXTENSION_get_critical(ext)
            data = backend._lib.X509_EXTENSION_get_data(ext)
            backend.openssl_assert(data != backend._ffi.NULL)
            der = backend._ffi.buffer(data.data, data.length)[:]
            entry = dict(
                critical=(crit == 1),
                value=to_native(base64.b64encode(der)),
            )
            try:
                oid = obj2txt(backend._lib, backend._ffi, backend._lib.X509_EXTENSION_get_object(ext))
            except AttributeError:
                oid = exts[i].oid.dotted_string
            result[oid] = entry

    except Exception:
        # In case the above method breaks, we likely have cryptography 36.0.0 or newer.
        # Use it's public_bytes() feature in that case. We will later switch this around
        # so that this code will be the default, but for now this will act as a fallback
        # since it will re-serialize de-serialized data, which can be different (if the
        # original data was not canonicalized) from what was contained in the CSR.
        for ext in csr.extensions:
            result[ext.oid.dotted_string] = dict(
                critical=ext.critical,
                value=to_native(base64.b64encode(ext.value.public_bytes())),
            )

    return result


def cryptography_name_to_oid(name):
    dotted = OID_LOOKUP.get(name)
    if dotted is None:
        if DOTTED_OID.match(name):
            return x509.oid.ObjectIdentifier(name)
        raise OpenSSLObjectError('Cannot find OID for "{0}"'.format(name))
    return x509.oid.ObjectIdentifier(dotted)


def cryptography_oid_to_name(oid, short=False):
    dotted_string = oid.dotted_string
    names = OID_MAP.get(dotted_string)
    if names:
        name = names[0]
    else:
        name = oid._name
        if name == 'Unknown OID':
            name = dotted_string
    if short:
        return NORMALIZE_NAMES_SHORT.get(name, name)
    else:
        return NORMALIZE_NAMES.get(name, name)


def _get_hex(bytesstr):
    if bytesstr is None:
        return bytesstr
    data = binascii.hexlify(bytesstr)
    data = to_text(b':'.join(data[i:i + 2] for i in range(0, len(data), 2)))
    return data


def _parse_hex(bytesstr):
    if bytesstr is None:
        return bytesstr
    data = ''.join([('0' * (2 - len(p)) + p) if len(p) < 2 else p for p in to_text(bytesstr).split(':')])
    data = binascii.unhexlify(data)
    return data


DN_COMPONENT_START_RE = re.compile(b'^ *([a-zA-z0-9.]+) *= *')
DN_HEX_LETTER = b'0123456789abcdef'


if sys.version_info[0] < 3:
    _int_to_byte = chr
else:
    def _int_to_byte(value):
        return bytes((value, ))


def _parse_dn_component(name, sep=b',', decode_remainder=True):
    m = DN_COMPONENT_START_RE.match(name)
    if not m:
        raise OpenSSLObjectError(u'cannot start part in "{0}"'.format(to_text(name)))
    oid = cryptography_name_to_oid(to_text(m.group(1)))
    idx = len(m.group(0))
    decoded_name = []
    sep_str = sep + b'\\'
    if decode_remainder:
        length = len(name)
        if length > idx and name[idx:idx + 1] == b'#':
            # Decoding a hex string
            idx += 1
            while idx + 1 < length:
                ch1 = name[idx:idx + 1]
                ch2 = name[idx + 1:idx + 2]
                idx1 = DN_HEX_LETTER.find(ch1.lower())
                idx2 = DN_HEX_LETTER.find(ch2.lower())
                if idx1 < 0 or idx2 < 0:
                    raise OpenSSLObjectError(u'Invalid hex sequence entry "{0}"'.format(to_text(ch1 + ch2)))
                idx += 2
                decoded_name.append(_int_to_byte(idx1 * 16 + idx2))
        else:
            # Decoding a regular string
            while idx < length:
                i = idx
                while i < length and name[i:i + 1] not in sep_str:
                    i += 1
                if i > idx:
                    decoded_name.append(name[idx:i])
                    idx = i
                while idx + 1 < length and name[idx:idx + 1] == b'\\':
                    ch = name[idx + 1:idx + 2]
                    idx1 = DN_HEX_LETTER.find(ch.lower())
                    if idx1 >= 0:
                        if idx + 2 >= length:
                            raise OpenSSLObjectError(u'Hex escape sequence "\\{0}" incomplete at end of string'.format(to_text(ch)))
                        ch2 = name[idx + 2:idx + 3]
                        idx2 = DN_HEX_LETTER.find(ch2.lower())
                        if idx2 < 0:
                            raise OpenSSLObjectError(u'Hex escape sequence "\\{0}" has invalid second letter'.format(to_text(ch + ch2)))
                        ch = _int_to_byte(idx1 * 16 + idx2)
                        idx += 1
                    idx += 2
                    decoded_name.append(ch)
                if idx < length and name[idx:idx + 1] == sep:
                    break
    else:
        decoded_name.append(name[idx:])
        idx = len(name)
    return x509.NameAttribute(oid, to_text(b''.join(decoded_name))), name[idx:]


def _parse_dn(name):
    '''
    Parse a Distinguished Name.

    Can be of the form ``CN=Test, O = Something`` or ``CN = Test,O= Something``.
    '''
    original_name = name
    name = name.lstrip()
    sep = b','
    if name.startswith(b'/'):
        sep = b'/'
        name = name[1:]
    result = []
    while name:
        try:
            attribute, name = _parse_dn_component(name, sep=sep)
        except OpenSSLObjectError as e:
            raise OpenSSLObjectError(u'Error while parsing distinguished name "{0}": {1}'.format(to_text(original_name), e))
        result.append(attribute)
        if name:
            if name[0:1] != sep or len(name) < 2:
                raise OpenSSLObjectError(u'Error while parsing distinguished name "{0}": unexpected end of string'.format(to_text(original_name)))
            name = name[1:]
    return result


def cryptography_parse_relative_distinguished_name(rdn):
    names = []
    for part in rdn:
        try:
            names.append(_parse_dn_component(to_bytes(part), decode_remainder=False)[0])
        except OpenSSLObjectError as e:
            raise OpenSSLObjectError(u'Error while parsing relative distinguished name "{0}": {1}'.format(part, e))
    return cryptography.x509.RelativeDistinguishedName(names)


def _is_ascii(value):
    '''Check whether the Unicode string `value` contains only ASCII characters.'''
    try:
        value.encode("ascii")
        return True
    except UnicodeEncodeError:
        return False


def _adjust_idn(value, idn_rewrite):
    if idn_rewrite == 'ignore' or not value:
        return value
    if idn_rewrite == 'idna' and _is_ascii(value):
        return value
    if idn_rewrite not in ('idna', 'unicode'):
        raise ValueError('Invalid value for idn_rewrite: "{0}"'.format(idn_rewrite))
    if not HAS_IDNA:
        raise OpenSSLObjectError(
            missing_required_lib('idna', reason='to transform {what} DNS name "{name}" to {dest}'.format(
                name=value,
                what='IDNA' if idn_rewrite == 'unicode' else 'Unicode',
                dest='Unicode' if idn_rewrite == 'unicode' else 'IDNA',
            )))
    # Since IDNA does not like '*' or empty labels (except one empty label at the end),
    # we split and let IDNA only handle labels that are neither empty or '*'.
    parts = value.split(u'.')
    for index, part in enumerate(parts):
        if part in (u'', u'*'):
            continue
        try:
            if idn_rewrite == 'idna':
                parts[index] = idna.encode(part).decode('ascii')
            elif idn_rewrite == 'unicode' and part.startswith(u'xn--'):
                parts[index] = idna.decode(part)
        except idna.IDNAError as exc2008:
            try:
                if idn_rewrite == 'idna':
                    parts[index] = part.encode('idna').decode('ascii')
                elif idn_rewrite == 'unicode' and part.startswith(u'xn--'):
                    parts[index] = part.encode('ascii').decode('idna')
            except Exception as exc2003:
                raise OpenSSLObjectError(
                    u'Error while transforming part "{part}" of {what} DNS name "{name}" to {dest}.'
                    u' IDNA2008 transformation resulted in "{exc2008}", IDNA2003 transformation resulted in "{exc2003}".'.format(
                        part=part,
                        name=value,
                        what='IDNA' if idn_rewrite == 'unicode' else 'Unicode',
                        dest='Unicode' if idn_rewrite == 'unicode' else 'IDNA',
                        exc2003=exc2003,
                        exc2008=exc2008,
                    ))
    return u'.'.join(parts)


def _adjust_idn_email(value, idn_rewrite):
    idx = value.find(u'@')
    if idx < 0:
        return value
    return u'{0}@{1}'.format(value[:idx], _adjust_idn(value[idx + 1:], idn_rewrite))


def _adjust_idn_url(value, idn_rewrite):
    url = urlparse(value)
    host = _adjust_idn(url.hostname, idn_rewrite)
    if url.username is not None and url.password is not None:
        host = u'{0}:{1}@{2}'.format(url.username, url.password, host)
    elif url.username is not None:
        host = u'{0}@{1}'.format(url.username, host)
    if url.port is not None:
        host = u'{0}:{1}'.format(host, url.port)
    return urlunparse(
        ParseResult(scheme=url.scheme, netloc=host, path=url.path, params=url.params, query=url.query, fragment=url.fragment))


def cryptography_get_name(name, what='Subject Alternative Name'):
    '''
    Given a name string, returns a cryptography x509.GeneralName object.
    Raises an OpenSSLObjectError if the name is unknown or cannot be parsed.
    '''
    try:
        if name.startswith('DNS:'):
            return x509.DNSName(_adjust_idn(to_text(name[4:]), 'idna'))
        if name.startswith('IP:'):
            address = to_text(name[3:])
            if '/' in address:
                return x509.IPAddress(ipaddress.ip_network(address))
            return x509.IPAddress(ipaddress.ip_address(address))
        if name.startswith('email:'):
            return x509.RFC822Name(_adjust_idn_email(to_text(name[6:]), 'idna'))
        if name.startswith('URI:'):
            return x509.UniformResourceIdentifier(_adjust_idn_url(to_text(name[4:]), 'idna'))
        if name.startswith('RID:'):
            m = re.match(r'^([0-9]+(?:\.[0-9]+)*)$', to_text(name[4:]))
            if not m:
                raise OpenSSLObjectError('Cannot parse {what} "{name}"'.format(name=name, what=what))
            return x509.RegisteredID(x509.oid.ObjectIdentifier(m.group(1)))
        if name.startswith('otherName:'):
            # otherName can either be a raw ASN.1 hex string or in the format that OpenSSL works with.
            m = re.match(r'^([0-9]+(?:\.[0-9]+)*);([0-9a-fA-F]{1,2}(?::[0-9a-fA-F]{1,2})*)$', to_text(name[10:]))
            if m:
                return x509.OtherName(x509.oid.ObjectIdentifier(m.group(1)), _parse_hex(m.group(2)))

            # See https://www.openssl.org/docs/man1.0.2/man5/x509v3_config.html - Subject Alternative Name for more
            # defailts on the format expected.
            name = to_text(name[10:], errors='surrogate_or_strict')
            if ';' not in name:
                raise OpenSSLObjectError('Cannot parse {what} otherName "{name}", must be in the '
                                         'format "otherName:<OID>;<ASN.1 OpenSSL Encoded String>" or '
                                         '"otherName:<OID>;<hex string>"'.format(name=name, what=what))

            oid, value = name.split(';', 1)
            b_value = serialize_asn1_string_as_der(value)
            return x509.OtherName(x509.ObjectIdentifier(oid), b_value)
        if name.startswith('dirName:'):
            return x509.DirectoryName(x509.Name(reversed(_parse_dn(to_bytes(name[8:])))))
    except Exception as e:
        raise OpenSSLObjectError('Cannot parse {what} "{name}": {error}'.format(name=name, what=what, error=e))
    if ':' not in name:
        raise OpenSSLObjectError('Cannot parse {what} "{name}" (forgot "DNS:" prefix?)'.format(name=name, what=what))
    raise OpenSSLObjectError('Cannot parse {what} "{name}" (potentially unsupported by cryptography backend)'.format(name=name, what=what))


def _dn_escape_value(value):
    '''
    Escape Distinguished Name's attribute value.
    '''
    value = value.replace(u'\\', u'\\\\')
    for ch in [u',', u'+', u'<', u'>', u';', u'"']:
        value = value.replace(ch, u'\\%s' % ch)
    value = value.replace(u'\0', u'\\00')
    if value.startswith((u' ', u'#')):
        value = u'\\%s' % value[0] + value[1:]
    if value.endswith(u' '):
        value = value[:-1] + u'\\ '
    return value


def cryptography_decode_name(name, idn_rewrite='ignore'):
    '''
    Given a cryptography x509.GeneralName object, returns a string.
    Raises an OpenSSLObjectError if the name is not supported.
    '''
    if idn_rewrite not in ('ignore', 'idna', 'unicode'):
        raise AssertionError('idn_rewrite must be one of "ignore", "idna", or "unicode"')
    if isinstance(name, x509.DNSName):
        return u'DNS:{0}'.format(_adjust_idn(name.value, idn_rewrite))
    if isinstance(name, x509.IPAddress):
        if isinstance(name.value, (ipaddress.IPv4Network, ipaddress.IPv6Network)):
            return u'IP:{0}/{1}'.format(name.value.network_address.compressed, name.value.prefixlen)
        return u'IP:{0}'.format(name.value.compressed)
    if isinstance(name, x509.RFC822Name):
        return u'email:{0}'.format(_adjust_idn_email(name.value, idn_rewrite))
    if isinstance(name, x509.UniformResourceIdentifier):
        return u'URI:{0}'.format(_adjust_idn_url(name.value, idn_rewrite))
    if isinstance(name, x509.DirectoryName):
        # According to https://datatracker.ietf.org/doc/html/rfc4514.html#section-2.1 the
        # list needs to be reversed, and joined by commas
        return u'dirName:' + ','.join([
            u'{0}={1}'.format(to_text(cryptography_oid_to_name(attribute.oid, short=True)), _dn_escape_value(attribute.value))
            for attribute in reversed(list(name.value))
        ])
    if isinstance(name, x509.RegisteredID):
        return u'RID:{0}'.format(name.value.dotted_string)
    if isinstance(name, x509.OtherName):
        return u'otherName:{0};{1}'.format(name.type_id.dotted_string, _get_hex(name.value))
    raise OpenSSLObjectError('Cannot decode name "{0}"'.format(name))


def _cryptography_get_keyusage(usage):
    '''
    Given a key usage identifier string, returns the parameter name used by cryptography's x509.KeyUsage().
    Raises an OpenSSLObjectError if the identifier is unknown.
    '''
    if usage in ('Digital Signature', 'digitalSignature'):
        return 'digital_signature'
    if usage in ('Non Repudiation', 'nonRepudiation'):
        return 'content_commitment'
    if usage in ('Key Encipherment', 'keyEncipherment'):
        return 'key_encipherment'
    if usage in ('Data Encipherment', 'dataEncipherment'):
        return 'data_encipherment'
    if usage in ('Key Agreement', 'keyAgreement'):
        return 'key_agreement'
    if usage in ('Certificate Sign', 'keyCertSign'):
        return 'key_cert_sign'
    if usage in ('CRL Sign', 'cRLSign'):
        return 'crl_sign'
    if usage in ('Encipher Only', 'encipherOnly'):
        return 'encipher_only'
    if usage in ('Decipher Only', 'decipherOnly'):
        return 'decipher_only'
    raise OpenSSLObjectError('Unknown key usage "{0}"'.format(usage))


def cryptography_parse_key_usage_params(usages):
    '''
    Given a list of key usage identifier strings, returns the parameters for cryptography's x509.KeyUsage().
    Raises an OpenSSLObjectError if an identifier is unknown.
    '''
    params = dict(
        digital_signature=False,
        content_commitment=False,
        key_encipherment=False,
        data_encipherment=False,
        key_agreement=False,
        key_cert_sign=False,
        crl_sign=False,
        encipher_only=False,
        decipher_only=False,
    )
    for usage in usages:
        params[_cryptography_get_keyusage(usage)] = True
    return params


def cryptography_get_basic_constraints(constraints):
    '''
    Given a list of constraints, returns a tuple (ca, path_length).
    Raises an OpenSSLObjectError if a constraint is unknown or cannot be parsed.
    '''
    ca = False
    path_length = None
    if constraints:
        for constraint in constraints:
            if constraint.startswith('CA:'):
                if constraint == 'CA:TRUE':
                    ca = True
                elif constraint == 'CA:FALSE':
                    ca = False
                else:
                    raise OpenSSLObjectError('Unknown basic constraint value "{0}" for CA'.format(constraint[3:]))
            elif constraint.startswith('pathlen:'):
                v = constraint[len('pathlen:'):]
                try:
                    path_length = int(v)
                except Exception as e:
                    raise OpenSSLObjectError('Cannot parse path length constraint "{0}" ({1})'.format(v, e))
            else:
                raise OpenSSLObjectError('Unknown basic constraint "{0}"'.format(constraint))
    return ca, path_length


def cryptography_key_needs_digest_for_signing(key):
    '''Tests whether the given private key requires a digest algorithm for signing.

    Ed25519 and Ed448 keys do not; they need None to be passed as the digest algorithm.
    '''
    if CRYPTOGRAPHY_HAS_ED25519 and isinstance(key, cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey):
        return False
    if CRYPTOGRAPHY_HAS_ED448 and isinstance(key, cryptography.hazmat.primitives.asymmetric.ed448.Ed448PrivateKey):
        return False
    return True


def _compare_public_keys(key1, key2, clazz):
    a = isinstance(key1, clazz)
    b = isinstance(key2, clazz)
    if not (a or b):
        return None
    if not a or not b:
        return False
    a = key1.public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw)
    b = key2.public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw)
    return a == b


def cryptography_compare_public_keys(key1, key2):
    '''Tests whether two public keys are the same.

    Needs special logic for Ed25519 and Ed448 keys, since they do not have public_numbers().
    '''
    if CRYPTOGRAPHY_HAS_ED25519:
        res = _compare_public_keys(key1, key2, cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey)
        if res is not None:
            return res
    if CRYPTOGRAPHY_HAS_ED448:
        res = _compare_public_keys(key1, key2, cryptography.hazmat.primitives.asymmetric.ed448.Ed448PublicKey)
        if res is not None:
            return res
    return key1.public_numbers() == key2.public_numbers()


def _compare_private_keys(key1, key2, clazz, has_no_private_bytes=False):
    a = isinstance(key1, clazz)
    b = isinstance(key2, clazz)
    if not (a or b):
        return None
    if not a or not b:
        return False
    if has_no_private_bytes:
        # We do not have the private_bytes() function - compare associated public keys
        return cryptography_compare_public_keys(a.public_key(), b.public_key())
    encryption_algorithm = cryptography.hazmat.primitives.serialization.NoEncryption()
    a = key1.private_bytes(serialization.Encoding.Raw, serialization.PrivateFormat.Raw, encryption_algorithm=encryption_algorithm)
    b = key2.private_bytes(serialization.Encoding.Raw, serialization.PrivateFormat.Raw, encryption_algorithm=encryption_algorithm)
    return a == b


def cryptography_compare_private_keys(key1, key2):
    '''Tests whether two private keys are the same.

    Needs special logic for Ed25519, X25519, and Ed448 keys, since they do not have private_numbers().
    '''
    if CRYPTOGRAPHY_HAS_ED25519:
        res = _compare_private_keys(key1, key2, cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey)
        if res is not None:
            return res
    if CRYPTOGRAPHY_HAS_X25519:
        res = _compare_private_keys(
            key1, key2, cryptography.hazmat.primitives.asymmetric.x25519.X25519PrivateKey, has_no_private_bytes=not CRYPTOGRAPHY_HAS_X25519_FULL)
        if res is not None:
            return res
    if CRYPTOGRAPHY_HAS_ED448:
        res = _compare_private_keys(key1, key2, cryptography.hazmat.primitives.asymmetric.ed448.Ed448PrivateKey)
        if res is not None:
            return res
    if CRYPTOGRAPHY_HAS_X448:
        res = _compare_private_keys(key1, key2, cryptography.hazmat.primitives.asymmetric.x448.X448PrivateKey)
        if res is not None:
            return res
    return key1.private_numbers() == key2.private_numbers()


def cryptography_serial_number_of_cert(cert):
    '''Returns cert.serial_number.

    Also works for old versions of cryptography.
    '''
    try:
        return cert.serial_number
    except AttributeError:
        # The property was called "serial" before cryptography 1.4
        return cert.serial


def parse_pkcs12(pkcs12_bytes, passphrase=None):
    '''Returns a tuple (private_key, certificate, additional_certificates, friendly_name).
    '''
    if _load_pkcs12 is None and _load_key_and_certificates is None:
        raise ValueError('neither load_pkcs12() nor load_key_and_certificates() present in the current cryptography version')

    if passphrase is not None:
        passphrase = to_bytes(passphrase)

    # Main code for cryptography 36.0.0 and forward
    if _load_pkcs12 is not None:
        return _parse_pkcs12_36_0_0(pkcs12_bytes, passphrase)

    if LooseVersion(cryptography.__version__) >= LooseVersion('35.0'):
        return _parse_pkcs12_35_0_0(pkcs12_bytes, passphrase)

    return _parse_pkcs12_legacy(pkcs12_bytes, passphrase)


def _parse_pkcs12_36_0_0(pkcs12_bytes, passphrase=None):
    # Requires cryptography 36.0.0 or newer
    pkcs12 = _load_pkcs12(pkcs12_bytes, passphrase)
    additional_certificates = [cert.certificate for cert in pkcs12.additional_certs]
    private_key = pkcs12.key
    certificate = None
    friendly_name = None
    if pkcs12.cert:
        certificate = pkcs12.cert.certificate
        friendly_name = pkcs12.cert.friendly_name
    return private_key, certificate, additional_certificates, friendly_name


def _parse_pkcs12_35_0_0(pkcs12_bytes, passphrase=None):
    # Backwards compatibility code for cryptography 35.x
    private_key, certificate, additional_certificates = _load_key_and_certificates(pkcs12_bytes, passphrase)

    friendly_name = None
    if certificate:
        # See https://github.com/pyca/cryptography/issues/5760#issuecomment-842687238
        backend = default_backend()

        # This code basically does what load_key_and_certificates() does, but without error-checking.
        # Since load_key_and_certificates succeeded, it should not fail.
        pkcs12 = backend._ffi.gc(
            backend._lib.d2i_PKCS12_bio(backend._bytes_to_bio(pkcs12_bytes).bio, backend._ffi.NULL),
            backend._lib.PKCS12_free)
        certificate_x509_ptr = backend._ffi.new("X509 **")
        with backend._zeroed_null_terminated_buf(to_bytes(passphrase) if passphrase is not None else None) as passphrase_buffer:
            backend._lib.PKCS12_parse(
                pkcs12,
                passphrase_buffer,
                backend._ffi.new("EVP_PKEY **"),
                certificate_x509_ptr,
                backend._ffi.new("Cryptography_STACK_OF_X509 **"))
        if certificate_x509_ptr[0] != backend._ffi.NULL:
            maybe_name = backend._lib.X509_alias_get0(certificate_x509_ptr[0], backend._ffi.NULL)
            if maybe_name != backend._ffi.NULL:
                friendly_name = backend._ffi.string(maybe_name)

    return private_key, certificate, additional_certificates, friendly_name


def _parse_pkcs12_legacy(pkcs12_bytes, passphrase=None):
    # Backwards compatibility code for cryptography < 35.0.0
    private_key, certificate, additional_certificates = _load_key_and_certificates(pkcs12_bytes, passphrase)

    friendly_name = None
    if certificate:
        # See https://github.com/pyca/cryptography/issues/5760#issuecomment-842687238
        backend = certificate._backend
        maybe_name = backend._lib.X509_alias_get0(certificate._x509, backend._ffi.NULL)
        if maybe_name != backend._ffi.NULL:
            friendly_name = backend._ffi.string(maybe_name)
    return private_key, certificate, additional_certificates, friendly_name


def cryptography_verify_signature(signature, data, hash_algorithm, signer_public_key):
    '''
    Check whether the given signature of the given data was signed by the given public key object.
    '''
    try:
        if CRYPTOGRAPHY_HAS_RSA_SIGN and isinstance(signer_public_key, cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey):
            signer_public_key.verify(signature, data, padding.PKCS1v15(), hash_algorithm)
            return True
        if CRYPTOGRAPHY_HAS_EC_SIGN and isinstance(signer_public_key, cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey):
            signer_public_key.verify(signature, data, cryptography.hazmat.primitives.asymmetric.ec.ECDSA(hash_algorithm))
            return True
        if CRYPTOGRAPHY_HAS_DSA_SIGN and isinstance(signer_public_key, cryptography.hazmat.primitives.asymmetric.dsa.DSAPublicKey):
            signer_public_key.verify(signature, data, hash_algorithm)
            return True
        if CRYPTOGRAPHY_HAS_ED25519_SIGN and isinstance(signer_public_key, cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PublicKey):
            signer_public_key.verify(signature, data)
            return True
        if CRYPTOGRAPHY_HAS_ED448_SIGN and isinstance(signer_public_key, cryptography.hazmat.primitives.asymmetric.ed448.Ed448PublicKey):
            signer_public_key.verify(signature, data)
            return True
        raise OpenSSLObjectError(u'Unsupported public key type {0}'.format(type(signer_public_key)))
    except InvalidSignature:
        return False


def cryptography_verify_certificate_signature(certificate, signer_public_key):
    '''
    Check whether the given X509 certificate object was signed by the given public key object.
    '''
    return cryptography_verify_signature(
        certificate.signature,
        certificate.tbs_certificate_bytes,
        certificate.signature_hash_algorithm,
        signer_public_key
    )



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

try:
    import boto3
    import botocore
except ImportError:
    pass  # will be captured by imported HAS_BOTO3

from ansible.plugins.inventory import BaseInventoryPlugin
from ansible.plugins.inventory import Cacheable
from ansible.plugins.inventory import Constructable

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.plugin_utils.base import AWSPluginBase
from ansible_collections.amazon.aws.plugins.plugin_utils.botocore import AnsibleBotocoreError


def _boto3_session(profile_name=None):
    if profile_name is None:
        return boto3.Session()
    return boto3.session.Session(profile_name=profile_name)


class AWSInventoryBase(BaseInventoryPlugin, Constructable, Cacheable, AWSPluginBase):
    class TemplatedOptions:
        # When someone looks up the TEMPLATABLE_OPTIONS using get() any templates
        # will be templated using the loader passed to parse.
        TEMPLATABLE_OPTIONS = (
            "access_key",
            "secret_key",
            "session_token",
            "profile",
            "iam_role_name",
        )

        def __init__(self, templar, options):
            self.original_options = options
            self.templar = templar

        def __getitem__(self, *args):
            return self.original_options.__getitem__(self, *args)

        def __setitem__(self, *args):
            return self.original_options.__setitem__(self, *args)

        def get(self, *args):
            value = self.original_options.get(*args)
            if not value:
                return value
            if args[0] not in self.TEMPLATABLE_OPTIONS:
                return value
            if not self.templar.is_template(value):
                return value

            return self.templar.template(variable=value, disable_lookups=False)

    def get_options(self, *args):
        original_options = super().get_options(*args)
        if not self.templar:
            return original_options
        return self.TemplatedOptions(self.templar, original_options)

    def __init__(self):
        super().__init__()
        self._frozen_credentials = {}

    # pylint: disable=too-many-arguments
    def parse(self, inventory, loader, path, cache=True, botocore_version=None, boto3_version=None):
        super().parse(inventory, loader, path)
        self.require_aws_sdk(botocore_version=botocore_version, boto3_version=boto3_version)
        self._read_config_data(path)
        self._set_frozen_credentials()

    def client(self, *args, **kwargs):
        kw_args = dict(self._frozen_credentials)
        kw_args.update(kwargs)
        return super().client(*args, **kw_args)

    def resource(self, *args, **kwargs):
        kw_args = dict(self._frozen_credentials)
        kw_args.update(kwargs)
        return super().resource(*args, **kw_args)

    def _freeze_iam_role(self, iam_role_arn):
        if hasattr(self, "ansible_name"):
            role_session_name = f"ansible_aws_{self.ansible_name}_dynamic_inventory"
        else:
            role_session_name = "ansible_aws_dynamic_inventory"
        assume_params = {"RoleArn": iam_role_arn, "RoleSessionName": role_session_name}

        try:
            sts = self.client("sts")
            assumed_role = sts.assume_role(**assume_params)
        except AnsibleBotocoreError as e:
            self.fail_aws(f"Unable to assume role {iam_role_arn}", exception=e)

        credentials = assumed_role.get("Credentials")
        if not credentials:
            self.fail_aws(f"Unable to assume role {iam_role_arn}")

        self._frozen_credentials = {
            "profile_name": None,
            "aws_access_key_id": credentials.get("AccessKeyId"),
            "aws_secret_access_key": credentials.get("SecretAccessKey"),
            "aws_session_token": credentials.get("SessionToken"),
        }

    def _set_frozen_credentials(self):
        options = self.get_options()
        iam_role_arn = options.get("assume_role_arn")
        if iam_role_arn:
            self._freeze_iam_role(iam_role_arn)

    def _describe_regions(self, service):
        # Try pulling a list of regions from the service
        try:
            initial_region = self.region or "us-east-1"
            client = self.client(service, region=initial_region)
            resp = client.describe_regions()
        except AttributeError:
            # Not all clients support describe
            pass
        except is_boto3_error_code("UnauthorizedOperation"):
            self.warn(f"UnauthorizedOperation when trying to list {service} regions")
        except botocore.exceptions.NoRegionError:
            self.warn(f"NoRegionError when trying to list {service} regions")
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            self.warn(f"Unexpected error while trying to list {service} regions: {e}")
        else:
            regions = [x["RegionName"] for x in resp.get("Regions", [])]
            if regions:
                return regions
        return None

    def _boto3_regions(self, service):
        options = self.get_options()

        if options.get("regions"):
            return options.get("regions")

        # boto3 has hard coded lists of available regions for resources, however this does bit-rot
        # As such we try to query the service, and fall back to ec2 for a list of regions
        for resource_type in list({service, "ec2"}):
            regions = self._describe_regions(resource_type)
            if regions:
                return regions

        # fallback to local list hardcoded in boto3 if still no regions
        session = _boto3_session(options.get("profile"))
        regions = session.get_available_regions(service)

        if not regions:
            # I give up, now you MUST give me regions
            self.fail_aws(
                "Unable to get regions list from available methods, you must specify the 'regions' option to continue."
            )

        return regions

    def all_clients(self, service):
        """
        Generator that yields a boto3 client and the region

        :param service: The boto3 service to connect to.

        Note: For services which don't support 'DescribeRegions' this may include bad
        endpoints, and as such EndpointConnectionError should be cleanly handled as a non-fatal
        error.
        """
        regions = self._boto3_regions(service=service)

        for region in regions:
            connection = self.client(service, region=region)
            yield connection, region

    def get_cached_result(self, path, cache):
        # false when refresh_cache or --flush-cache is used
        if not cache:
            return False, None
        # get the user-specified directive
        if not self.get_option("cache"):
            return False, None

        cache_key = self.get_cache_key(path)
        try:
            cached_value = self._cache[cache_key]
        except KeyError:
            # if cache expires or cache file doesn"t exist
            return False, None

        return True, cached_value

    def update_cached_result(self, path, cache, result):
        if not self.get_option("cache"):
            return

        cache_key = self.get_cache_key(path)
        # We weren't explicitly told to flush the cache, and there's already a cache entry,
        # this means that the result we're being passed came from the cache.  As such we don't
        # want to "update" the cache as that could reset a TTL on the cache entry.
        if cache and cache_key in self._cache:
            return

        self._cache[cache_key] = result

    def verify_file(self, path):
        """
        :param path: the path to the inventory config file
        :return the contents of the config file
        """
        if not super().verify_file(path):
            return False

        if hasattr(self, "INVENTORY_FILE_SUFFIXES"):
            if not path.endswith(self.INVENTORY_FILE_SUFFIXES):
                return False

        return True



#!/usr/bin/python
#
# Copyright (c) 2021 Cole Neubauer, (@coleneubauer)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
module: azure_rm_adgroup
version_added: "1.6.0"
short_description: Manage Azure Active Directory group
description:
    - Create, update or delete Azure Active Directory group.
options:
    tenant:
        description:
            - The tenant ID.
        type: str
        required: True
    state:
        description:
            - Assert the state of the resource group. Use C(present) to create or update and C(absent) to delete.
        default: present
        choices:
            - absent
            - present
        type: str
    object_id:
        description:
            - The object id for the ad group.
            - Can be used to reference when updating an existing group.
            - Ignored when attempting to create a group.
        type: str
    display_name:
        description:
            - The display name of the ad group.
            - Can be used with I(mail_nickname) instead of I(object_id) to reference existing group.
            - Required when creating a new ad group.
        type: str
    mail_nickname:
        description:
            - The mail nickname of the ad group.
            - Can be used with I(display_name) instead of I(object_id) to reference existing group.
            - Required when creating a new ad group.
        type: str
    present_members:
        description:
            - The azure ad objects asserted to be members of the group.
            - This list does not need to be all inclusive. Objects that are members and not on this list remain members.
        type: list
        elements: str
    absent_members:
        description:
            - The azure ad objects asserted to not be members of the group.
        type: list
        elements: str
    present_owners:
        description:
            - The azure ad objects asserted to be owners of the group.
            - This list does not need to be all inclusive. Objects that are owners and not on this list remain members.
        type: list
        elements: str
    absent_owners:
        description:
            - The azure ad objects asserted to not be owners of the group.
        type: list
        elements: str
extends_documentation_fragment:
    - azure.azcollection.azure
author:
    - Cole Neubauer(@coleneubauer)
'''

EXAMPLES = '''
- name: Create Group
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    display_name: "Group-Name"
    mail_nickname: "Group-Mail-Nickname"
    state: 'present'

- name: Delete Group using display_name and mail_nickname
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    display_name: "Group-Name"
    mail_nickname: "Group-Mail-Nickname"
    state: 'absent'

- name: Delete Group using object_id
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    object_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    state: 'absent'

- name: Ensure Users are Members of a Group using display_name and mail_nickname
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    display_name: "Group-Name"
    mail_nickname: "Group-Mail-Nickname"
    state: 'present'
    present_members:
      - "https://graph.windows.net/{{ tenant_id }}/directoryObjects/{{ ad_object_1_object_id }}"
      - "https://graph.windows.net/{{ tenant_id }}/directoryObjects/{{ ad_object_2_object_id }}"

- name: Ensure Users are Members of a Group using object_id
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    object_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    state: 'present'
    present_members:
      - "https://graph.windows.net/{{ ad_object_1_tenant_id }}/directoryObjects/{{ ad_object_1_object_id }}"
      - "https://graph.windows.net/{{ ad_object_2_tenant_id }}/directoryObjects/{{ ad_object_2_object_id }}"

- name: Ensure Users are not Members of a Group using display_name and mail_nickname
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    display_name: "Group-Name"
    mail_nickname: "Group-Mail-Nickname"
    state: 'present'
    absent_members:
      - "{{ ad_object_1_object_id }}"

- name: Ensure Users are Members of a Group using object_id
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    object_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    state: 'present'
    absent_members:
      - "{{ ad_object_1_object_id }}"

- name: Ensure Users are Owners of a Group using display_name and mail_nickname
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    display_name: "Group-Name"
    mail_nickname: "Group-Mail-Nickname"
    state: 'present'
    present_owners:
      - "https://graph.windows.net/{{ tenant_id }}/directoryObjects/{{ ad_object_1_object_id }}"
      - "https://graph.windows.net/{{ tenant_id }}/directoryObjects/{{ ad_object_2_object_id }}"

- name: Ensure Users are Owners of a Group using object_id
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    object_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    state: 'present'
    present_owners:
      - "https://graph.windows.net/{{ ad_object_1_tenant_id }}/directoryObjects/{{ ad_object_1_object_id }}"
      - "https://graph.windows.net/{{ ad_object_2_tenant_id }}/directoryObjects/{{ ad_object_2_object_id }}"

- name: Ensure Users are not Owners of a Group using display_name and mail_nickname
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    display_name: "Group-Name"
    mail_nickname: "Group-Mail-Nickname"
    state: 'present'
    absent_owners:
      - "{{ ad_object_1_object_id }}"
      - "{{ ad_object_2_object_id }}"

- name: Ensure Users are Owners of a Group using object_id
  azure_rm_adgroup:
    tenant: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    object_id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    state: 'present'
    absent_owners:
      - "{{ ad_object_1_object_id }}"
      - "{{ ad_object_2_object_id }}"
'''

RETURN = '''
object_id:
    description:
        - The object_id for the group.
    type: str
    returned: always
    sample: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
display_name:
    description:
        - The display name of the group.
    returned: always
    type: str
    sample: GroupName
mail_nickname:
    description:
        - The mail alias for the group.
    returned: always
    type: str
    sample: groupname
mail_enabled:
    description:
        - Whether the group is mail-enabled. Must be false. This is because only pure security groups can be created using the Graph API.
    returned: always
    type: bool
    sample: False
security_enabled:
    description:
        - Whether the group is security-enable.
    returned: always
    type: bool
    sample: False
mail:
    description:
        - The primary email address of the group.
    returned: always
    type: str
    sample: group@contoso.com
group_owners:
    description:
        - The owners of the group.
    returned: always
    type: list
group_members:
    description:
        - The members of the group.
    returned: always
    type: list
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBase

try:
    from azure.graphrbac.models import GraphErrorException
    from azure.graphrbac.models import GroupCreateParameters
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMADGroup(AzureRMModuleBase):
    def __init__(self):

        self.module_arg_spec = dict(
            object_id=dict(type='str'),
            display_name=dict(type='str'),
            mail_nickname=dict(type='str'),
            present_members=dict(type='list', elements='str'),
            present_owners=dict(type='list', elements='str'),
            absent_members=dict(type='list', elements='str'),
            absent_owners=dict(type='list', elements='str'),
            tenant=dict(type='str', required=True),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            ),
        )

        self.tenant = None
        self.display_name = None
        self.mail_nickname = None
        self.object_id = None
        self.present_members = []
        self.present_owners = []
        self.absent_members = []
        self.absent_owners = []
        self.state = None
        self.results = dict(changed=False)

        super(AzureRMADGroup, self).__init__(derived_arg_spec=self.module_arg_spec,
                                             supports_check_mode=False,
                                             supports_tags=False,
                                             is_ad_resource=True)

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        # TODO remove ad_groups return. Returns as one object always
        ad_groups = []

        try:
            client = self.get_graphrbac_client(self.tenant)
            ad_groups = []

            if self.display_name and self.mail_nickname:
                ad_groups = list(client.groups.list(filter="displayName eq '{0}' and mailNickname eq '{1}'".format(self.display_name, self.mail_nickname)))

                if ad_groups:
                    self.object_id = ad_groups[0].object_id

            elif self.object_id:
                ad_groups = [client.groups.get(self.object_id)]

            if ad_groups:
                if self.state == "present":
                    self.results["changed"] = False
                elif self.state == "absent":
                    ad_groups = [client.groups.delete(self.object_id)]
                    self.results["changed"] = True
            else:
                if self.state == "present":
                    if self.display_name and self.mail_nickname:
                        ad_groups = [client.groups.create(GroupCreateParameters(display_name=self.display_name, mail_nickname=self.mail_nickname))]
                        self.results["changed"] = True
                    else:
                        raise ValueError('The group does not exist. Both display_name : {0} and mail_nickname : {1} must be passed to create a new group'
                                         .format(self.display_name, self.mail_nickname))
                elif self.state == "absent":
                    self.results["changed"] = False

            if ad_groups[0] is not None:
                self.update_members(ad_groups[0].object_id, client)
                self.update_owners(ad_groups[0].object_id, client)
                self.results.update(self.set_results(ad_groups[0], client))

        except GraphErrorException as e:
            self.fail(e)
        except ValueError as e:
            self.fail(e)

        return self.results

    def update_members(self, group_id, client):

        current_members = []

        if self.present_members or self.absent_members:
            current_members = [object.object_id for object in list(client.groups.get_group_members(group_id))]

        if self.present_members:
            present_members_by_object_id = self.dictionary_from_object_urls(self.present_members)

            members_to_add = list(set(present_members_by_object_id.keys()) - set(current_members))

            if members_to_add:
                for member_object_id in members_to_add:
                    client.groups.add_member(group_id, present_members_by_object_id[member_object_id])

                self.results["changed"] = True

        if self.absent_members:
            members_to_remove = list(set(self.absent_members).intersection(set(current_members)))

            if members_to_remove:
                for member in members_to_remove:
                    client.groups.remove_member(group_id, member)
                self.results["changed"] = True

    def update_owners(self, group_id, client):
        current_owners = []

        if self.present_owners or self.absent_owners:
            current_owners = [object.object_id for object in list(client.groups.list_owners(group_id))]

        if self.present_owners:

            present_owners_by_object_id = self.dictionary_from_object_urls(self.present_owners)

            owners_to_add = list(set(present_owners_by_object_id.keys()) - set(current_owners))

            if owners_to_add:
                for owner_object_id in owners_to_add:
                    client.groups.add_owner(group_id, present_owners_by_object_id[owner_object_id])
                self.results["changed"] = True

        if self.absent_owners:
            owners_to_remove = list(set(self.absent_owners).intersection(set(current_owners)))

            if owners_to_remove:
                for owner in owners_to_remove:
                    client.groups.remove_owner(group_id, owner)
                self.results["changed"] = True

    def dictionary_from_object_urls(self, object_urls):
        objects_by_object_id = {}

        for urls in object_urls:
            object_id = urls.split("/")[-1]
            objects_by_object_id[object_id] = urls

        return objects_by_object_id

    def application_to_dict(self, object):
        return dict(
            app_id=object.app_id,
            object_id=object.object_id,
            display_name=object.display_name,
        )

    def serviceprincipal_to_dict(self, object):
        return dict(
            app_id=object.app_id,
            object_id=object.object_id,
            app_display_name=object.display_name,
            app_role_assignment_required=object.app_role_assignment_required
        )

    def group_to_dict(self, object):
        return dict(
            object_id=object.object_id,
            display_name=object.display_name,
            mail_nickname=object.mail_nickname,
            mail_enabled=object.mail_enabled,
            security_enabled=object.security_enabled,
            mail=object.mail
        )

    def user_to_dict(self, object):
        return dict(
            object_id=object.object_id,
            display_name=object.display_name,
            user_principal_name=object.user_principal_name,
            mail_nickname=object.mail_nickname,
            mail=object.mail,
            account_enabled=object.account_enabled,
            user_type=object.user_type
        )

    def result_to_dict(self, object):
        if object.object_type == "Group":
            return self.group_to_dict(object)
        elif object.object_type == "User":
            return self.user_to_dict(object)
        elif object.object_type == "Application":
            return self.application_to_dict(object)
        elif object.object_type == "ServicePrincipal":
            return self.serviceprincipal_to_dict(object)
        else:
            return object.object_type

    def set_results(self, object, client):
        results = self.group_to_dict(object)

        if results["object_id"] and (self.present_owners or self.absent_owners):
            results["group_owners"] = [self.result_to_dict(object) for object in list(client.groups.list_owners(results["object_id"]))]

        if results["object_id"] and (self.present_members or self.absent_members):
            results["group_members"] = [self.result_to_dict(object) for object in list(client.groups.get_group_members(results["object_id"]))]

        return results


def main():
    AzureRMADGroup()


if __name__ == '__main__':
    main()
